# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Test Structured units and quantities specifically with the ERFA ufuncs.
"""

import numpy as np
import pytest
from erfa import ufunc as erfa_ufunc
from numpy.testing import assert_array_equal

from astropy import units as u
from astropy.tests.helper import assert_quantity_allclose
from astropy.units.quantity_helper.erfa import astrom_unit

ASTROM_UNIT = astrom_unit()


def vvd(val, valok, dval, func, test, status):
    """Mimic routine of erfa/src/t_erfa_c.c (to help copy & paste)"""
    assert u.allclose(val, valok * val.unit, atol=dval * val.unit)


class TestPVUfuncs:
    @classmethod
    def setup_class(cls):
        cls.pv_unit = u.Unit("AU,AU/day")
        cls.pv_value = np.array(
            [
                ([1.0, 0.0, 0.0], [0.0, 0.0125, 0.0]),
                ([0.0, 1.0, 0.0], [-0.0125, 0.0, 0.0]),
            ],
            dtype=erfa_ufunc.dt_pv,
        )
        cls.pv = cls.pv_value << cls.pv_unit

    def test_cpv(self):
        pv_copy = erfa_ufunc.cpv(self.pv)
        assert_array_equal(pv_copy, self.pv)
        assert not np.may_share_memory(pv_copy, self.pv)

    def test_p2pv(self):
        p2pv = erfa_ufunc.p2pv(self.pv["p"])
        assert_array_equal(p2pv["p"], self.pv["p"])
        assert_array_equal(
            p2pv["v"], np.zeros(self.pv.shape + (3,), float) << u.m / u.s
        )

    def test_p2pv_inplace(self):
        # TODO: fix np.zeros_like.
        out = np.zeros_like(self.pv_value) << self.pv_unit
        p2pv = erfa_ufunc.p2pv(self.pv["p"], out=out)
        assert out is p2pv
        assert_array_equal(p2pv["p"], self.pv["p"])
        assert_array_equal(
            p2pv["v"], np.zeros(self.pv.shape + (3,), float) << u.m / u.s
        )

    def test_pv2p(self):
        p = erfa_ufunc.pv2p(self.pv)
        assert_array_equal(p, self.pv["p"])
        out = np.zeros_like(p)
        p2 = erfa_ufunc.pv2p(self.pv, out=out)
        assert out is p2
        assert_array_equal(p2, self.pv["p"])

    def test_pv2s(self):
        theta, phi, r, td, pd, rd = erfa_ufunc.pv2s(self.pv)
        assert theta.unit == u.radian
        assert_quantity_allclose(theta, [0, 90] * u.deg)  # longitude
        assert phi.unit == u.radian
        assert_array_equal(phi.value, np.zeros(self.pv.shape))  # latitude
        assert r.unit == u.AU
        assert_array_equal(r.value, np.ones(self.pv.shape))
        assert td.unit == u.radian / u.day
        assert_array_equal(td.value, np.array([0.0125] * 2))
        assert pd.unit == u.radian / u.day
        assert_array_equal(pd.value, np.zeros(self.pv.shape))
        assert rd.unit == u.AU / u.day
        assert_array_equal(rd.value, np.zeros(self.pv.shape))

    def test_pv2s_non_standard_units(self):
        pv = self.pv_value << u.Unit("Pa,Pa/m")
        theta, phi, r, td, pd, rd = erfa_ufunc.pv2s(pv)
        assert theta.unit == u.radian
        assert_quantity_allclose(theta, [0, 90] * u.deg)  # longitude
        assert phi.unit == u.radian
        assert_array_equal(phi.value, np.zeros(pv.shape))  # latitude
        assert r.unit == u.Pa
        assert_array_equal(r.value, np.ones(pv.shape))
        assert td.unit == u.radian / u.m
        assert_array_equal(td.value, np.array([0.0125] * 2))
        assert pd.unit == u.radian / u.m
        assert_array_equal(pd.value, np.zeros(pv.shape))
        assert rd.unit == u.Pa / u.m
        assert_array_equal(rd.value, np.zeros(pv.shape))

    @pytest.mark.xfail(
        reason=(
            "erfa ufuncs cannot take different names; it is not yet clear whether "
            "this is changeable; see https://github.com/liberfa/pyerfa/issues/77"
        )
    )
    def test_pv2s_non_standard_names_and_units(self):
        pv_value = np.array(self.pv_value, dtype=[("pos", "f8"), ("vel", "f8")])
        pv = pv_value << u.Unit("Pa,Pa/m")
        theta, phi, r, td, pd, rd = erfa_ufunc.pv2s(pv)
        assert theta.unit == u.radian
        assert_quantity_allclose(theta, [0, 90] * u.deg)  # longitude
        assert phi.unit == u.radian
        assert_array_equal(phi.value, np.zeros(pv.shape))  # latitude
        assert r.unit == u.Pa
        assert_array_equal(r.value, np.ones(pv.shape))
        assert td.unit == u.radian / u.m
        assert_array_equal(td.value, np.array([0.0125] * 2))
        assert pd.unit == u.radian / u.m
        assert_array_equal(pd.value, np.zeros(pv.shape))
        assert rd.unit == u.Pa / u.m
        assert_array_equal(rd.value, np.zeros(pv.shape))

    def test_s2pv(self):
        theta, phi, r, td, pd, rd = erfa_ufunc.pv2s(self.pv)
        # On purpose change some of the units away from expected by s2pv.
        pv = erfa_ufunc.s2pv(
            theta.to(u.deg), phi, r.to(u.m), td.to(u.deg / u.day), pd, rd.to(u.m / u.s)
        )
        assert pv.unit == u.StructuredUnit("m, m/s", names=("p", "v"))
        assert_quantity_allclose(pv["p"], self.pv["p"], atol=1 * u.m, rtol=0)
        assert_quantity_allclose(pv["v"], self.pv["v"], atol=1 * u.mm / u.s, rtol=0)

    def test_s2p_not_all_quantity(self):
        # Test for a useful error message - see gh-16873.
        # Non-quantity input should be treated as dimensionless and thus cannot
        # be converted to radians.
        with pytest.raises(
            AttributeError,
            match=(
                "'NoneType' object has no attribute 'get_converter'"
                ".*\n.*treated as dimensionless"
            ),
        ):
            erfa_ufunc.s2p(0.5, 0.5, 4 * u.km)

        # Except if we have the right equivalency in place.
        with u.add_enabled_equivalencies(u.dimensionless_angles()):
            result = erfa_ufunc.s2p(0.5, 0.5, 4 * u.km)

        expected = erfa_ufunc.s2p(0.5 * u.radian, 0.5 * u.radian, 4 * u.km)
        assert_array_equal(result, expected)

    def test_pvstar(self):
        ra, dec, pmr, pmd, px, rv, stat = erfa_ufunc.pvstar(self.pv)
        assert_array_equal(stat, np.zeros(self.pv.shape, dtype="i4"))
        assert ra.unit == u.radian
        assert_quantity_allclose(ra, [0, 90] * u.deg)
        assert dec.unit == u.radian
        assert_array_equal(dec.value, np.zeros(self.pv.shape))  # latitude
        assert pmr.unit == u.radian / u.year
        assert_quantity_allclose(pmr, [0.0125, 0.0125] * u.radian / u.day)
        assert pmd.unit == u.radian / u.year
        assert_array_equal(pmd.value, np.zeros(self.pv.shape))
        assert px.unit == u.arcsec
        assert_quantity_allclose(px, 1 * u.radian)
        assert rv.unit == u.km / u.s
        # RV is non-zero because proper motion induces a small redshift
        # due to second order Doppler shift.
        assert_quantity_allclose(
            rv, np.zeros(self.pv.shape) << (u.km / u.s), atol=1 * u.m / u.s
        )

    def test_starpv(self):
        ra, dec, pmr, pmd, px, rv, stat = erfa_ufunc.pvstar(self.pv)
        pv, stat = erfa_ufunc.starpv(
            ra.to(u.deg), dec.to(u.deg), pmr, pmd, px, rv.to(u.m / u.s)
        )
        assert_array_equal(stat, np.zeros(self.pv.shape, dtype="i4"))
        assert pv.unit == self.pv.unit
        # Roundtrip is not as good as hoped on 32bit, not clear why.
        # But proper motions are ridiculously high...
        assert_quantity_allclose(pv["p"], self.pv["p"], atol=1 * u.m, rtol=0)
        assert_quantity_allclose(pv["v"], self.pv["v"], atol=1 * u.m / u.s, rtol=0)

    def test_pvtob(self):
        pv = erfa_ufunc.pvtob(
            [90, 0] * u.deg,
            0.0 * u.deg,
            100 * u.km,
            0 * u.deg,
            0 * u.deg,
            0 * u.deg,
            90 * u.deg,
        )
        assert pv.unit == u.StructuredUnit("m, m/s", names=("p", "v"))
        assert pv.unit["v"] == u.m / u.s
        assert_quantity_allclose(
            pv["p"], [[-6478, 0, 0], [0, 6478, 0]] * u.km, atol=2 * u.km
        )
        assert_quantity_allclose(
            pv["v"], [[0, -0.5, 0], [-0.5, 0, 0]] * u.km / u.s, atol=0.1 * u.km / u.s
        )

    def test_pvdpv(self):
        pvdpv = erfa_ufunc.pvdpv(self.pv, self.pv)
        assert pvdpv["pdp"].unit == self.pv.unit["p"] ** 2
        assert pvdpv["pdv"].unit == self.pv.unit["p"] * self.pv.unit["v"]
        assert_array_equal(
            pvdpv["pdp"], np.einsum("...i,...i->...", self.pv["p"], self.pv["p"])
        )
        assert_array_equal(
            pvdpv["pdv"], 2 * np.einsum("...i,...i->...", self.pv["p"], self.pv["v"])
        )
        z_axis = u.Quantity(np.array(([0, 0, 1], [0, 0, 0]), erfa_ufunc.dt_pv), "1,1/s")
        pvdpv2 = erfa_ufunc.pvdpv(self.pv, z_axis)
        assert pvdpv2["pdp"].unit == self.pv.unit["p"]
        assert pvdpv2["pdv"].unit == self.pv.unit["v"]
        assert_array_equal(pvdpv2["pdp"].value, np.zeros(self.pv.shape))
        assert_array_equal(pvdpv2["pdv"].value, np.zeros(self.pv.shape))

    def test_pvxpv(self):
        pvxpv = erfa_ufunc.pvxpv(self.pv, self.pv)
        assert pvxpv["p"].unit == self.pv.unit["p"] ** 2
        assert pvxpv["v"].unit == self.pv.unit["p"] * self.pv.unit["v"]
        assert_array_equal(pvxpv["p"].value, np.zeros(self.pv["p"].shape))
        assert_array_equal(pvxpv["v"].value, np.zeros(self.pv["v"].shape))
        z_axis = u.Quantity(np.array(([0, 0, 1], [0, 0, 0]), erfa_ufunc.dt_pv), "1,1/s")
        pvxpv2 = erfa_ufunc.pvxpv(self.pv, z_axis)
        assert pvxpv2["p"].unit == self.pv.unit["p"]
        assert pvxpv2["v"].unit == self.pv.unit["v"]
        assert_array_equal(pvxpv2["p"], [[0.0, -1, 0.0], [1.0, 0.0, 0.0]] * u.AU)
        assert_array_equal(
            pvxpv2["v"], [[0.0125, 0.0, 0.0], [0.0, 0.0125, 0.0]] * u.AU / u.day
        )

    def test_pvm(self):
        pm, vm = erfa_ufunc.pvm(self.pv)
        assert pm.unit == self.pv.unit["p"]
        assert vm.unit == self.pv.unit["v"]
        assert_array_equal(pm, np.linalg.norm(self.pv["p"], axis=-1))
        assert_array_equal(vm, np.linalg.norm(self.pv["v"], axis=-1))

    def test_pvmpv(self):
        pvmpv = erfa_ufunc.pvmpv(self.pv, self.pv)
        assert pvmpv.unit == self.pv.unit
        assert_array_equal(pvmpv["p"], 0 * self.pv["p"])
        assert_array_equal(pvmpv["v"], 0 * self.pv["v"])

    def test_pvppv(self):
        pvppv = erfa_ufunc.pvppv(self.pv, self.pv)
        assert pvppv.unit == self.pv.unit
        assert_array_equal(pvppv["p"], 2 * self.pv["p"])
        assert_array_equal(pvppv["v"], 2 * self.pv["v"])

    def test_pvu(self):
        pvu = erfa_ufunc.pvu(86400 * u.s, self.pv)
        assert pvu.unit == self.pv.unit
        assert_array_equal(pvu["p"], self.pv["p"] + 1 * u.day * self.pv["v"])
        assert_array_equal(pvu["v"], self.pv["v"])

    def test_pvup(self):
        pvup = erfa_ufunc.pvup(86400 * u.s, self.pv)
        assert pvup.unit == self.pv.unit["p"]
        assert_array_equal(pvup, self.pv["p"] + 1 * u.day * self.pv["v"])

    def test_sxpv(self):
        # Not a realistic example!!
        sxpv = erfa_ufunc.sxpv(10.0, self.pv)
        assert sxpv.unit == self.pv.unit
        assert_array_equal(sxpv["p"], self.pv["p"] * 10)
        assert_array_equal(sxpv["v"], self.pv["v"] * 10)
        sxpv2 = erfa_ufunc.sxpv(30.0 * u.s, self.pv)
        assert sxpv2.unit == u.StructuredUnit("AU s,AU s/d", names=("p", "v"))
        assert_array_equal(sxpv2["p"], self.pv["p"] * 30 * u.s)
        assert_array_equal(sxpv2["v"], self.pv["v"] * 30 * u.s)

    def test_s2xpv(self):
        # Not a realistic example!!
        s2xpv = erfa_ufunc.s2xpv(10.0, 1 * u.s, self.pv)
        assert s2xpv.unit == u.StructuredUnit("AU,AU s/d", names=("p", "v"))
        assert_array_equal(s2xpv["p"], self.pv["p"] * 10)
        assert_array_equal(s2xpv["v"], self.pv["v"] * u.s)

    @pytest.mark.parametrize(
        "r",
        [
            np.eye(3),
            np.array(
                [
                    [0.0, -1.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0],
                ]
            ),
            np.eye(3) / u.s,
        ],
    )
    def test_rxpv(self, r):
        result = erfa_ufunc.rxpv(r, self.pv)
        assert_array_equal(result["p"], np.einsum("...ij,...j->...i", r, self.pv["p"]))
        assert_array_equal(result["v"], np.einsum("...ij,...j->...i", r, self.pv["v"]))

    @pytest.mark.parametrize(
        "r",
        [
            np.eye(3),
            np.array(
                [
                    [0.0, -1.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0],
                ]
            ),
            np.eye(3) / u.s,
        ],
    )
    def test_trxpv(self, r):
        result = erfa_ufunc.trxpv(r, self.pv)
        assert_array_equal(
            result["p"], np.einsum("...ij,...j->...i", r.T, self.pv["p"])
        )
        assert_array_equal(
            result["v"], np.einsum("...ij,...j->...i", r.T, self.pv["v"])
        )


class TestEraStructUfuncs:
    @classmethod
    def setup_class(cls):
        ldbody = np.array(
            [
                (0.00028574, 3e-10, ([-7.81014427, -5.60956681, -1.98079819],
                                     [0.0030723249, -0.00406995477, -0.00181335842])),
                (0.00095435, 3e-9, ([0.738098796, 4.63658692, 1.9693136],
                                    [-0.00755816922, 0.00126913722, 0.000727999001])),
                (1.0, 6e-6, ([-0.000712174377, -0.00230478303, -0.00105865966],
                             [6.29235213e-6, -3.30888387e-7, -2.96486623e-7]))
            ],
            dtype=erfa_ufunc.dt_eraLDBODY
        )  # fmt: skip
        ldbody_unit = u.StructuredUnit("Msun,radian,(AU,AU/day)", ldbody.dtype)
        cls.ldbody = ldbody << ldbody_unit
        cls.ob = [-0.974170437, -0.2115201, -0.0917583114] << u.AU
        cls.sc = np.array([-0.763276255, -0.608633767, -0.216735543])

        # From t_atciq in t_erfa_c.c
        astrom, eo = erfa_ufunc.apci13(2456165.5, 0.401182685)
        cls.astrom = astrom << ASTROM_UNIT
        cls.rc = 2.71 * u.rad
        cls.dc = 0.174 * u.rad
        cls.pr = 1e-5 * u.rad / u.year
        cls.pd = 5e-6 * u.rad / u.year
        cls.px = 0.1 * u.arcsec
        cls.rv = 55.0 * u.km / u.s

    def test_ldn_basic(self):
        sn = erfa_ufunc.ldn(self.ldbody, self.ob, self.sc)
        assert_quantity_allclose(
            sn,
            [-0.7632762579693333866, -0.6086337636093002660, -0.2167355420646328159]
            * u.one,
            atol=1e-12,
            rtol=0,
        )

    def test_ldn_in_other_unit(self):
        ldbody = self.ldbody.to("kg,rad,(m,m/s)")
        ob = self.ob.to("m")
        sn = erfa_ufunc.ldn(ldbody, ob, self.sc)
        assert_quantity_allclose(
            sn,
            [-0.7632762579693333866, -0.6086337636093002660, -0.2167355420646328159]
            * u.one,
            atol=1e-12,
            rtol=0,
        )

    def test_ldn_in_SI(self):
        sn = erfa_ufunc.ldn(self.ldbody.si, self.ob.si, self.sc)
        assert_quantity_allclose(
            sn,
            [-0.7632762579693333866, -0.6086337636093002660, -0.2167355420646328159]
            * u.one,
            atol=1e-12,
            rtol=0,
        )

    def test_aper(self):
        along = self.astrom["along"]
        astrom2 = erfa_ufunc.aper(10 * u.deg, self.astrom)
        assert astrom2["eral"].unit == u.radian
        assert_quantity_allclose(astrom2["eral"], along + 10 * u.deg)
        astrom3 = self.astrom.to("s,km,1,km,1,1,1,deg,deg,deg,deg,1,1,1,rad,rad,rad")
        astrom4 = erfa_ufunc.aper(10 * u.deg, astrom3)
        assert astrom3["eral"].unit == u.rad
        assert astrom4["eral"].unit == u.deg
        assert astrom4.unit == "s,km,1,km,1,1,1,deg,deg,deg,deg,1,1,1,deg,rad,rad"
        assert_quantity_allclose(astrom4["eral"], along + 10 * u.deg)

    def test_atciq_basic(self):
        ri, di = erfa_ufunc.atciq(
            self.rc, self.dc, self.pr, self.pd, self.px, self.rv, self.astrom
        )
        assert_quantity_allclose(ri, 2.710121572968696744 * u.rad)
        assert_quantity_allclose(di, 0.1729371367219539137 * u.rad)

    def test_atciq_in_other_unit(self):
        astrom = self.astrom.to("s,km,1,km,1,1,1,deg,deg,deg,deg,1,1,1,deg,deg,deg")
        ri, di = erfa_ufunc.atciq(
            self.rc.to(u.deg),
            self.dc.to(u.deg),
            self.pr.to(u.mas / u.yr),
            self.pd.to(u.mas / u.yr),
            self.px,
            self.rv.to(u.m / u.s),
            astrom,
        )
        assert_quantity_allclose(ri, 2.710121572968696744 * u.rad, atol=1e-12 * u.rad)
        assert_quantity_allclose(di, 0.1729371367219539137 * u.rad, atol=1e-12 * u.rad)

    def test_atciqn(self):
        ri, di = erfa_ufunc.atciqn(
            self.rc.to(u.deg),
            self.dc.to(u.deg),
            self.pr.to(u.mas / u.yr),
            self.pd.to(u.mas / u.yr),
            self.px,
            self.rv.to(u.m / u.s),
            self.astrom.si,
            self.ldbody.si,
        )
        assert_quantity_allclose(ri, 2.710122008104983335 * u.rad, atol=1e-12 * u.rad)
        assert_quantity_allclose(di, 0.1729371916492767821 * u.rad, atol=1e-12 * u.rad)

    def test_atciqz(self):
        ri, di = erfa_ufunc.atciqz(self.rc.to(u.deg), self.dc.to(u.deg), self.astrom.si)
        assert_quantity_allclose(ri, 2.709994899247256984 * u.rad, atol=1e-12 * u.rad)
        assert_quantity_allclose(di, 0.1728740720984931891 * u.rad, atol=1e-12 * u.rad)

    def test_aticq(self):
        ri = 2.710121572969038991 * u.rad
        di = 0.1729371367218230438 * u.rad
        rc, dc = erfa_ufunc.aticq(ri.to(u.deg), di.to(u.deg), self.astrom.si)
        assert_quantity_allclose(rc, 2.710126504531716819 * u.rad, atol=1e-12 * u.rad)
        assert_quantity_allclose(dc, 0.1740632537627034482 * u.rad, atol=1e-12 * u.rad)

    def test_aticqn(self):
        ri = 2.709994899247599271 * u.rad
        di = 0.1728740720983623469 * u.rad
        rc, dc = erfa_ufunc.aticqn(
            ri.to(u.deg), di.to(u.deg), self.astrom.si, self.ldbody.si
        )
        assert_quantity_allclose(rc, 2.709999575033027333 * u.rad, atol=1e-12 * u.rad)
        assert_quantity_allclose(dc, 0.1739999656316469990 * u.rad, atol=1e-12 * u.rad)

    def test_atioq_atoiq(self):
        astrom, _ = erfa_ufunc.apio13(
            2456384.5,
            0.969254051,
            0.1550675,
            -0.527800806,
            -1.2345856,
            2738.0,
            2.47230737e-7,
            1.82640464e-6,
            731.0,
            12.8,
            0.59,
            0.55,
        )
        astrom = astrom << ASTROM_UNIT

        ri = 2.710121572969038991 * u.rad
        di = 0.1729371367218230438 * u.rad
        aob, zob, hob, dob, rob = erfa_ufunc.atioq(
            ri.to(u.deg), di.to(u.deg), astrom.si
        )
        assert_quantity_allclose(
            aob, 0.9233952224895122499e-1 * u.rad, atol=1e-12 * u.rad
        )
        assert_quantity_allclose(zob, 1.407758704513549991 * u.rad, atol=1e-12 * u.rad)
        assert_quantity_allclose(
            hob, -0.9247619879881698140e-1 * u.rad, atol=1e-12 * u.rad
        )
        assert_quantity_allclose(dob, 0.1717653435756234676 * u.rad, atol=1e-12 * u.rad)
        assert_quantity_allclose(rob, 2.710085107988480746 * u.rad, atol=1e-12 * u.rad)

        # Sadly does not just use the values from above.
        ob1 = 2.710085107986886201 * u.rad
        ob2 = 0.1717653435758265198 * u.rad
        ri2, di2 = erfa_ufunc.atoiq("R", ob1.to(u.deg), ob2.to(u.deg), astrom.si)
        assert_quantity_allclose(ri2, 2.710121574447540810 * u.rad, atol=1e-12 * u.rad)
        assert_quantity_allclose(
            di2, 0.17293718391166087785 * u.rad, atol=1e-12 * u.rad
        )


class TestAp:
    @classmethod
    def setup_class(cls):
        # Values from apco13 and apio test cases in t_erfa_c.c
        # with units changed so we can check conversion.
        cls.utc1 = 2456384.5
        cls.utc2 = 0.969254051
        cls.dut1 = (0.1550675 * u.s).to(u.ms)
        cls.sp = (-3.01974337e-11 * u.rad).to(u.deg)
        cls.theta = (3.14540971 * u.rad).to(u.mdeg)
        cls.elong = (-0.527800806 * u.rad).to(u.Marcsec)
        cls.phi = (-1.2345856 * u.rad).to(u.arcmin)
        cls.hm = (2738.0 * u.m).to(u.km)
        cls.phpa = (731.0 * u.hPa).to(u.Pa)
        cls.tc = (12.8 * u.deg_C).to(u.K, equivalencies=u.temperature())
        cls.rh = (0.59 * u.one).to(u.percent)
        cls.wl = (0.55 * u.micron).to(u.AA)
        # For apio.
        cls.xp = (2.47230737e-7 * u.rad).to(u.arcsec)
        cls.yp = (1.82640464e-6 * u.rad).to(u.arcmin)
        cls.refa = (0.000201418779 * u.rad).to(u.mas)
        cls.refb = (-2.36140831e-7 * u.rad).to(u.uas)
        cls.apco13_args = [
            getattr(cls, name)
            for name in [
                "utc1",
                "utc2",
                "dut1",
                "elong",
                "phi",
                "hm",
                "xp",
                "yp",
                "phpa",
                "tc",
                "rh",
                "wl",
            ]
        ]
        cls.apio_args = [
            getattr(cls, name)
            for name in [
                "sp",
                "theta",
                "elong",
                "phi",
                "hm",
                "xp",
                "yp",
                "refa",
                "refb",
            ]
        ]

    def test_apco13(self):
        astrom, eo, status = erfa_ufunc.apco13(*self.apco13_args)
        assert status == 0
        vvd(eo, -0.003020548354802412839, 1e-14, "eraApco13", "eo", status)
        assert astrom.unit == ASTROM_UNIT
        for name, expected in {
            "pmt": 13.25248468622475727,
            "eb": [-0.9741827107320875162,
                   -0.2115130190489716682,
                   -0.09179840189496755339],
            "eh": [-0.9736425572586935247,
                   -0.2092452121603336166,
                   -0.09075578153885665295],
            "em": 0.9998233240913898141,
            "v": [0.2078704994520489246e-4,
                  -0.8955360133238868938e-4,
                  -0.3863338993055887398e-4],
            "bm1": 0.9999999950277561004,
            "bpn": np.array(
                [[ 0.9999991390295147999,
                   0.4978650075315529277e-7,
                   0.001312227200850293372],
                 [-0.1136336652812486604e-7,
                   0.9999999995713154865,
                  -0.2928086230975367296e-4],
                 [-0.001312227201745553566,
                   0.2928082218847679162e-4,
                   0.9999991386008312212],
                ]).T,
            "along": -0.5278008060295995733,
            "xpl": 0.1133427418130752958e-5,
            "ypl": 0.1453347595780646207e-5,
            "sphi": -0.9440115679003211329,
            "cphi": 0.3299123514971474711,
            "diurab": 0,
            "eral": 2.617608909189664000,
            "refa": 0.2014187785940396921e-3,
            "refb": -0.2361408314943696227e-6,
        }.items():  # fmt: skip
            assert_quantity_allclose(astrom[name], expected * ASTROM_UNIT[name])

    def test_apco13_no_time_units(self):
        msg = "cannot pass in units for 2-part time in apco13"
        with pytest.raises(TypeError, match=msg):
            erfa_ufunc.apco13(self.utc1 * u.day, *self.apco13_args[1:])

        with pytest.raises(TypeError, match=msg):
            erfa_ufunc.apco13(self.utc1, self.utc2 * u.day, *self.apco13_args[2:])

        with pytest.raises(TypeError, match=msg):
            erfa_ufunc.apco13(
                self.utc1 * u.day, self.utc2 * u.day, *self.apco13_args[2:]
            )

    def test_apio(self):
        astrom = erfa_ufunc.apio(*self.apio_args)
        assert astrom.unit == ASTROM_UNIT
        for name, expected in {
            "along": -0.5278008060295995734,
            "xpl": 0.1133427418130752958e-5,
            "ypl": 0.1453347595780646207e-5,
            "sphi": -0.9440115679003211329,
            "cphi": 0.3299123514971474711,
            "diurab": 0.5135843661699913529e-6,
            "eral": 2.617608903970400427,
            "refa": 0.2014187790000000000e-3,
            "refb": -0.2361408310000000000e-6,
        }.items():
            assert_quantity_allclose(astrom[name], expected * ASTROM_UNIT[name])


class TestGeodetic:
    @classmethod
    def setup_class(cls):
        cls.ellipsoid = 1
        cls.length_unit = u.Unit("m")
        cls.equatorial_radius_value = 6378136.0
        cls.equatorial_radius = (cls.equatorial_radius_value << cls.length_unit).to(
            u.km
        )
        cls.flattening = 0.0033528 * u.dimensionless_unscaled
        cls.lon_value = 0.9827937232473290680
        cls.lon_unit = u.Unit("rad")
        cls.lon = (cls.lon_value << cls.lon_unit).to(u.deg)
        cls.lat_value = 0.9716018377570411532
        cls.lat_unit = u.Unit("rad")
        cls.lat = (cls.lat_value << cls.lat_unit).to(u.deg)
        cls.height_value = 332.36862495764397
        cls.height = cls.height_value << cls.length_unit
        cls.x_value = 2e6
        cls.x = (cls.x_value << cls.length_unit).to(u.cm)
        cls.y_value = 3e6
        cls.y = (cls.y_value << cls.length_unit).to(u.km)
        cls.z_value = 5.244e6
        cls.z = (cls.z_value << cls.length_unit).to(u.Mm)
        cls.xyz = np.stack([cls.x, cls.y, cls.z])

    def test_unit_errors(self):
        """Test unit errors when dimensionless parameters are used"""

        msg = (
            "'NoneType' object has no attribute 'get_converter'"
            ".*\n.*treated as dimensionless"
        )
        with pytest.raises(AttributeError, match=msg):
            erfa_ufunc.gc2gde(self.equatorial_radius_value, self.flattening, self.xyz)
        with pytest.raises(AttributeError, match=msg):
            erfa_ufunc.gd2gce(
                self.equatorial_radius_value,
                self.flattening,
                self.lon,
                self.lat,
                self.height,
            )
        with pytest.raises(AttributeError, match=msg):
            erfa_ufunc.gc2gde(self.equatorial_radius, self.flattening, self.xyz.value)
        with pytest.raises(AttributeError, match=msg):
            erfa_ufunc.gd2gce(
                self.equatorial_radius,
                self.flattening,
                self.lon_value,
                self.lat,
                self.height,
            )
        with pytest.raises(AttributeError, match=msg):
            erfa_ufunc.gd2gce(
                self.equatorial_radius,
                self.flattening,
                self.lon,
                self.lat,
                self.height_value,
            )

    def test_gc2gde(self):
        """Test that we reproduce erfa/src/t_erfa_c.c t_gc2gd"""

        e, p, h, status = erfa_ufunc.gc2gde(
            self.equatorial_radius, self.flattening, self.xyz
        )
        assert status == 0
        vvd(e, self.lon_value, 1e-14, "eraGc2gde", "e", status)
        vvd(p, self.lat_value, 1e-14, "eraGc2gde", "p", status)
        vvd(h, self.height_value, 1e-8, "eraGc2gde", "h", status)

    def test_gd2gce(self):
        """Test that we reproduce erfa/src/t_erfa_c.c t_gc2gd"""

        xyz, status = erfa_ufunc.gd2gce(
            self.equatorial_radius, self.flattening, self.lon, self.lat, self.height
        )
        assert status == 0
        vvd(xyz[0], self.x_value, 1e-7, "eraGd2gce", "e", status)
        vvd(xyz[1], self.y_value, 1e-7, "eraGd2gce", "p", status)
        vvd(xyz[2], self.z_value, 1e-7, "eraGd2gce", "h", status)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Some might be indirectly tested already in ``astropy.io.fits.tests``.
"""

import io

import numpy as np
import pytest

from astropy.table import Table
from astropy.utils.diff import diff_values, report_diff_values, where_not_allclose


@pytest.mark.parametrize("a", [np.nan, np.inf, 1.11, 1, "a"])
def test_diff_values_false(a):
    assert not diff_values(a, a)


@pytest.mark.parametrize(
    ("a", "b"),
    [(np.inf, np.nan), (1.11, 1.1), (1, 2), (1, "a"), ("a", "b"), (np.nan, 1.1)],
)
def test_diff_values_true(a, b):
    assert diff_values(a, b)


def test_float_comparison():
    """
    Regression test for https://github.com/spacetelescope/PyFITS/issues/21
    """
    f = io.StringIO()
    a = np.float32(0.029751372)
    b = np.float32(0.029751368)
    identical = report_diff_values(a, b, fileobj=f)
    assert not identical
    out = f.getvalue()

    # This test doesn't care about what the exact output is, just that it
    # did show a difference in their text representations
    assert "a>" in out
    assert "b>" in out


def test_diff_types():
    """
    Regression test for https://github.com/astropy/astropy/issues/4122
    """
    f = io.StringIO()
    a = 1.0
    b = "1.0"
    identical = report_diff_values(a, b, fileobj=f)
    assert not identical
    assert f.getvalue().splitlines() == [
        "  (float) a> 1.0",
        "    (str) b> '1.0'",
        "           ? +   +",
    ]


def test_diff_numeric_scalar_types():
    """Test comparison of different numeric scalar types."""
    f = io.StringIO()
    assert not report_diff_values(1.0, 1, fileobj=f)
    out = f.getvalue()
    assert out == "  (float) a> 1.0\n    (int) b> 1\n"


def test_array_comparison():
    """
    Test diff-ing two arrays.
    """
    f = io.StringIO()
    a = np.arange(9).reshape(3, 3)
    b = a + 1
    identical = report_diff_values(a, b, fileobj=f)
    assert not identical
    out = f.getvalue()
    assert (
        out == "  at [0, 0]:\n"
        "    a> 0\n"
        "    b> 1\n"
        "  at [0, 1]:\n"
        "    a> 1\n"
        "    b> 2\n"
        "  at [0, 2]:\n"
        "    a> 2\n"
        "    b> 3\n"
        "  ...and at 6 more indices.\n"
    )


def test_diff_shaped_array_comparison():
    """
    Test diff-ing two differently shaped arrays.
    """
    f = io.StringIO()
    a = np.empty((1, 2, 3))
    identical = report_diff_values(a, a[0], fileobj=f)
    assert not identical
    out = f.getvalue()
    assert (
        out
        == "  Different array shapes:\n    a> (1, 2, 3)\n     ?  ---\n    b> (2, 3)\n"
    )


def test_tablediff():
    """
    Test diff-ing two simple Table objects.
    """
    a = Table.read(
        """name    obs_date    mag_b  mag_v
M31     2012-01-02  17.0   16.0
M82     2012-10-29  16.2   15.2
M101    2012-10-31  15.1   15.5""",
        format="ascii",
    )
    b = Table.read(
        """name    obs_date    mag_b  mag_v
M31     2012-01-02  17.0   16.5
M82     2012-10-29  16.2   15.2
M101    2012-10-30  15.1   15.5
NEW     2018-05-08   nan    9.0""",
        format="ascii",
    )
    f = io.StringIO()
    identical = report_diff_values(a, b, fileobj=f)
    assert not identical
    out = f.getvalue()
    assert (
        out == "     name  obs_date  mag_b mag_v\n"
        "     ---- ---------- ----- -----\n"
        "  a>  M31 2012-01-02  17.0  16.0\n"
        "   ?                           ^\n"
        "  b>  M31 2012-01-02  17.0  16.5\n"
        "   ?                           ^\n"
        "      M82 2012-10-29  16.2  15.2\n"
        "  a> M101 2012-10-31  15.1  15.5\n"
        "   ?               ^\n"
        "  b> M101 2012-10-30  15.1  15.5\n"
        "   ?               ^\n"
        "  b>  NEW 2018-05-08   nan   9.0\n"
    )

    # Identical
    assert report_diff_values(a, a, fileobj=f)


def test_large_table_diff():
    # see https://github.com/astropy/astropy/issues/14010
    colnames = [f"column{i}" for i in range(100)]
    t1 = Table(names=colnames)

    colnames.insert(50, "test")
    t2 = Table(names=colnames)

    assert not report_diff_values(t1, t2, fileobj=io.StringIO())


@pytest.mark.parametrize("kwargs", [{}, {"atol": 0, "rtol": 0}])
def test_where_not_allclose(kwargs):
    a = np.array([1, np.nan, np.inf, 4.5])
    b = np.array([1, np.inf, np.nan, 4.6])

    assert where_not_allclose(a, b, **kwargs) == ([3],)
    assert len(where_not_allclose(a, a, **kwargs)[0]) == 0

    diff, maxabs, maxrel = where_not_allclose(a, b, return_maxdiff=True, **kwargs)
    assert np.isclose(maxabs, 0.1)
    assert np.isclose(maxrel, 0.1 / 4.6)

    a = np.array([np.nan, np.inf, 0, 4.5])
    b = np.array([1, np.inf, np.nan, 4.6])

    diff, maxabs, maxrel = where_not_allclose(a, b, return_maxdiff=True, **kwargs)
    assert list(diff[0]) == [0, 2, 3]
    assert np.isclose(maxabs, 0.1)
    assert np.isclose(maxrel, 0.1 / 4.6)

    a = np.array([np.nan, np.inf, 0])
    b = np.array([1, np.inf, np.nan])

    diff, maxabs, maxrel = where_not_allclose(a, b, return_maxdiff=True, **kwargs)
    assert list(diff[0]) == [0, 2]
    assert maxabs == 0
    assert maxrel == 0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""This module contains helper functions and classes for handling metadata."""

from . import core as _core
from . import exceptions as _exceptions
from . import merge as _merge
from . import utils as _utils
from .core import *
from .exceptions import *
from .merge import *
from .utils import *

__all__ = []
__all__ += _core.__all__
__all__ += _merge.__all__
__all__ += _exceptions.__all__
__all__ += _utils.__all__



from astropy.io.fits.hdu.table import BinTableHDU

__all__ = ["_CompBinTableHDU"]


class _CompBinTableHDU(BinTableHDU):
    _load_variable_length_data = False
    """
    We don't want to always load all the tiles so by setting this option
    we can then access the tiles as needed.
    """

    _manages_own_heap = True

    @classmethod
    def match_header(cls, header):
        return False



# Licensed under a 3-clause BSD style license - see LICENSE.rst


from astropy import table
from astropy.table import pprint


class MyRow(table.Row):
    def __str__(self):
        return str(self.as_void())


class MyColumn(table.Column):
    pass


class MyMaskedColumn(table.MaskedColumn):
    pass


class MyTableColumns(table.TableColumns):
    pass


class MyTableFormatter(pprint.TableFormatter):
    pass


class MyTable(table.Table):
    Row = MyRow
    Column = MyColumn
    MaskedColumn = MyMaskedColumn
    TableColumns = MyTableColumns
    TableFormatter = MyTableFormatter


def test_simple_subclass():
    t = MyTable([[1, 2], [3, 4]])
    row = t[0]
    assert isinstance(row, MyRow)
    assert isinstance(t["col0"], MyColumn)
    assert isinstance(t.columns, MyTableColumns)
    assert isinstance(t.formatter, MyTableFormatter)

    t2 = MyTable(t)
    row = t2[0]
    assert isinstance(row, MyRow)
    assert str(row) == "(1, 3)"

    t3 = table.Table(t)
    row = t3[0]
    assert not isinstance(row, MyRow)
    assert str(row) != "(1, 3)"

    t = MyTable([[1, 2], [3, 4]], masked=True)
    row = t[0]
    assert isinstance(row, MyRow)
    assert str(row) == "(1, 3)"
    assert isinstance(t["col0"], MyMaskedColumn)
    assert isinstance(t.formatter, MyTableFormatter)


class ParamsRow(table.Row):
    """
    Row class that allows access to an arbitrary dict of parameters
    stored as a dict object in the ``params`` column.
    """

    def __getitem__(self, item):
        if item not in self.colnames:
            return super().__getitem__("params")[item]
        else:
            return super().__getitem__(item)

    def keys(self):
        out = [name for name in self.colnames if name != "params"]
        params = [key.lower() for key in sorted(self["params"])]
        return out + params

    def values(self):
        return [self[key] for key in self.keys()]


class ParamsTable(table.Table):
    Row = ParamsRow


def test_params_table():
    t = ParamsTable(names=["a", "b", "params"], dtype=["i", "f", "O"])
    t.add_row((1, 2.0, {"x": 1.5, "y": 2.5}))
    t.add_row((2, 3.0, {"z": "hello", "id": 123123}))
    assert t["params"][0] == {"x": 1.5, "y": 2.5}
    assert t[0]["params"] == {"x": 1.5, "y": 2.5}
    assert t[0]["y"] == 2.5
    assert t[1]["id"] == 123123
    assert list(t[1].keys()) == ["a", "b", "id", "z"]
    assert list(t[1].values()) == [2, 3.0, 123123, "hello"]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""This package contains functions for reading and writing TDAT tables that are
not meant to be used directly, but instead are available as readers/writers in
`astropy.table`. See :ref:`astropy:table_io` for more details.

:Author: Daniel Giles (daniel.k.giles@gmail.com)
         Abdu Zoghbi

"""

import re
from copy import deepcopy
from warnings import warn

import numpy as np

from astropy.utils.exceptions import AstropyWarning

from . import basic, core

_STD_MSG = "See details in https://heasarc.gsfc.nasa.gov/docs/software/dbdocs/tdat.html"


def make_example_data():
    return [
        "<HEADER>",
        "# # and // are comments",
        "table_name = example_table",
        'table_description = "Example table"',
        "#",
        "# Table Parameters",
        "#",
        "field[id] = integer [meta.id] (key) // Unique ID",
        "field[ra] = float:.4f_degree [pos.eq.ra] (index) // Right Ascension",
        "field[name] = char12 [meta.id] // Name",
        "#",
        "# Virtual Parameters",
        "#",
        "table_author = Example et al.",
        "#",
        "# Data Format Specification",
        "#",
        "line[1] = id name ra",
        "<DATA>",
        "1|TargetOne|1.0|",
        "2|TargetTwo|2.0|",
        "<END>",
    ]


class TdatFormatError(Exception):
    """Tdat Format Error"""

    def __str__(self):
        return f"{super().__str__()}\n{_STD_MSG}"


class TdatFormatWarning(AstropyWarning):
    """Tdat Format Warning"""


class TdatDataSplitter(core.BaseSplitter):
    """Splitter for tdat data.

    Handles the (deprecated) cases of multiple data delimiters, record
    delimiters, and multi-line records.
    Multiple data delimiters - Multiple delimiters can be specified in the
        header, e.g. field_delimiter = "|!" would treat both | and !
        as individual delimiters. Default: "|"
    Record Delimiters - The record_delimiter can be specified in the header. By
        default there is no record delimiter and new records should be set on
        new lines. The following list standard escaped character sequences and their
        equivalent meanings can be used:
        * \t (tab)
        * \b (backspace)
        * \r (carriage return)
        * \f (form feed)
        * \v (vertical tab)
        * \a (audible alert/bell),
        * \\### (where ### is a number between 1 and 127 and represents the
        ASCII character with that numerical code).
        Note: Specifying a record delimiter value of "" is interpreted as a
            single blank line between records.
    Multi-line records - A single record may take more than one line, indicated in the header

    """

    delimiter = "|"
    record_delimiter = None

    @property
    def literals_dict(self):
        """Return a dictionary of placeholders to be used in place of
        backslashed delimiter characters.
        """
        return {c: f"literal{n}" for n, c in enumerate(self.delimiter)}

    def preprocess_data_lines(self, lines, record_delimiter):
        """Split lines into multiple lines if a record_delimiter is specified."""
        data_lines = []
        for line in lines:
            data_lines += re.split(rf"{record_delimiter}", line)
        return data_lines

    def process_line(self, line: str) -> str:
        """Remove whitespace at the beginning or end of line.  This is especially useful for
        whitespace-delimited files to prevent spurious columns at the beginning or end.

        READ: override default to handle backslashed delimiter characters.
        """
        for c in self.delimiter:
            line = line.replace(f"\\{c}", self.literals_dict[c])
        line = re.sub(rf"[{self.delimiter}]", "|", line)
        return line.strip()

    def __call__(self, lines, field_delimiter="|", record_delimiter=None, nlines=1):
        """ """
        self.delimiter = field_delimiter

        def replace_placeholders(line):
            for c in field_delimiter:
                line = line.replace(self.literals_dict[c], c)
            return line

        if " " in field_delimiter:
            warn(
                TdatFormatWarning("Double check your data when using space delimiters.")
            )

        if record_delimiter is not None:
            lines = self.preprocess_data_lines(lines, record_delimiter)
        if hasattr(self, "process_line"):
            lines = (self.process_line(x) for x in lines)

        iline = 0
        _lines = []
        for line in lines:
            _lines.append(line)
            iline += 1
            if iline == nlines:
                vals = [
                    replace_placeholders(val)
                    for _line in _lines
                    for val in re.split(r"\|", _line)[:-1]
                ]
                # Reset
                iline = 0
                _lines = []
                if hasattr(self, "process_val"):
                    yield [self.process_val(x) for x in vals]
                else:  # pragma: no cover
                    yield vals
            else:  # pragma: no cover
                continue

    def join(self, vals):
        delimiter = getattr(self, "delimiter", "|")
        # TDAT specification requires a delimiter at the end of each record
        return delimiter.join(str(x) for x in vals) + delimiter


class TdatHeader(basic.BasicHeader):
    """TDAT table header"""

    splitter_class = TdatDataSplitter
    cols = None
    comment = r"\s*(#|//)"
    # comments: # or //
    _comment = r"\s*(#|//)(.*)$"
    # descriptors that shouldn't be registered as comments
    _desc_comments = [
        "Table Parameters",
        "Virtual Parameters",
        "Relationship Definitions",
        "Data Format Specification",
    ]
    # keywords in the header: name = value
    _keys = r"(?P<key>\w+)\s*=\s*(?P<value>.*?)(\s*(#|//)(?P<comment>.*$)|(\s*$))"
    # keywords in the header: name[text] = some_other_text;
    # names: relate|line
    _extra_keys = r"\s*(relate|line)\[(\w+)\]\s*=\s*([\w\s]+)(?:\((\w+)\))?"
    _field_line = r"\s*field\[\w+\]\s*=\.*"

    _deprecated_keys = r"\s*(record_delimiter|field_delimiter)\s*=\s\"(.+)\""
    _required_keywords = ("table_name",)
    _dtype_dict_in = {
        "int1": np.int8,
        "integer1": np.int8,
        "tinyint": np.int8,
        "int2": np.int16,
        "integer2": np.int16,
        "smallint": np.int16,
        "int4": np.int32,
        "integer4": np.int32,
        "integer": np.int32,
        "float4": np.float32,
        "real": np.float32,
        "float": float,
        "float8": float,
    }

    _dtype_dict_out = {
        "int": "int4",
        "int32": "int4",
        "int16": "int2",
        "int8": "int1",
        "float": "float8",
        "float64": "float8",
        "float32": "float4",
    }

    def _validate_comment(self, line) -> bool:
        """Check if line is a valid comment, return comment if so"""
        comment_parser = re.compile(self._comment)
        cmatch = comment_parser.match(line)
        if cmatch:
            # Ignore common section headers
            for cc in self._desc_comments:
                if cc in cmatch.group(2):
                    return False
            # Ignore empty comments
            return cmatch.group(2) != ""
        return False

    def _process_comment(self, line: str) -> str:
        return re.sub("^" + self.comment, "", line).strip()

    def _process_keywords(self, lines):
        line_fields = {}
        keywords = getattr(self, "_keywords", {})
        keys_parser = re.compile(self._keys)
        extra_keys_parser = re.compile(self._extra_keys)
        deprecated_keys_parser = re.compile(self._deprecated_keys)

        for line in lines:
            kmatch = keys_parser.match(line)
            ematch = extra_keys_parser.match(line)
            dmatch = deprecated_keys_parser.match(line)
            if dmatch:
                warn(
                    f'"{dmatch.group(1)}" keyword is deprecated. {_STD_MSG}',
                    TdatFormatWarning,
                )
                if dmatch.group(1) == "record_delimiter":
                    self.record_delimiter = dmatch.group(2)
                elif dmatch.group(1) == "field_delimiter":
                    self.delimiter = dmatch.group(2)
            elif ematch:
                # match extra keywords
                if ematch.group(1) == "relate":
                    warn(
                        f'"relate" keyword is obsolete and will be ignored. {_STD_MSG}',
                        TdatFormatWarning,
                    )
                elif ematch.group(1) == "line":
                    # fields in line; typically only 1, but can be more
                    line_fields[ematch.group(2)] = ematch.group(3).split()

            elif kmatch:
                key = kmatch.group("key")
                _quotes = r"^(?P<quote>[\"\'\`]*)(?P<value>.*?)(?P<endquote>[\"\'\`]*$)"
                quote_parse = re.compile(_quotes)
                quotes_match = quote_parse.match(kmatch.group("value"))
                if quotes_match.group("quote") != quotes_match.group("endquote")[::-1]:
                    raise TdatFormatError("Mismatched quotes for value of " + key)

                value = kmatch.group("value").strip("\"'`")
                if key == "table_name":
                    if len(value) > 20:
                        warn(
                            "The table_name has a maximum length of 20 characters, truncating.",
                            TdatFormatWarning,
                        )
                        value = value[:20]
                elif key == "table_description":
                    if len(value) > 80:
                        warn(
                            "The table_description has a maximum length of 80 characters, truncating.",
                            TdatFormatWarning,
                        )
                        value = value[:80]
                elif key == "table_security":
                    if value.lower() not in ("public", "private"):
                        warn(
                            "Value for table_security not recognized, should be public|private.",
                            TdatFormatWarning,
                        )
                keywords[key] = value

        self._line_fields = line_fields

        if len(keywords) > len(getattr(self, "_keywords", {})):
            self._keywords = keywords

    def update_meta(self, lines, meta):
        """Extract meta information: comments and key/values in the header
        READ: Overrides the default update_meta
        """
        try:
            start_line = (
                min(
                    i
                    for i, line in enumerate(lines)
                    if line.strip().upper() == "<HEADER>"
                )
                + 1
            )
        except ValueError:
            raise TdatFormatError("<HEADER> not found in file." + _STD_MSG)
        try:
            end_line = min(
                i for i, line in enumerate(lines) if line.strip().upper() == "<DATA>"
            )
        except ValueError:
            raise TdatFormatError("<DATA> not found in file." + _STD_MSG)

        meta["table"]["comments"] = [
            self._process_comment(line)
            for line in lines[start_line:end_line]
            if self._validate_comment(line)
        ]
        self._process_keywords(lines)
        meta["table"]["keywords"] = self._keywords

    def process_lines(self, lines):
        """Select and process lines between <HEADER> and <DATA> lines
        READ: Override default process_lines
        """
        fl_parser = re.compile(self._field_line)
        start_line = (
            min(i for i, line in enumerate(lines) if line.strip().upper() == "<HEADER>")
            + 1
        )
        end_line = min(
            i for i, line in enumerate(lines) if line.strip().upper() == "<DATA>"
        )

        for line in lines[start_line:end_line]:
            if fl_parser.match(line):
                yield line

    def get_cols(self, lines):
        """Initialize the header Column objects from TDAT field lines
        READ: Overrides default get_cols
        """
        col_parser = re.compile(
            r"""
            \s*field
            \[(?P<name>\w+)\]\s*=\s*
            (?P<ctype>[^\W\s_]+(\(\d+\))?)
            (?:\:(?P<fmt>[^_\s\[\(/#]+))?
            (?:_(?P<unit>[^\[\(#]+))?\s*
            (?:\[(?P<ucd>[\w\.\;]+)\])?
            \s*
            (?:\((?P<index>\w+)\))?
            \s*
            (?:(//|\#)\s*(?P<desc>[^/#]*))?
            (?:(//|\#)\s*(?P<comment>.*))?
            \s*
            """,
            re.VERBOSE,
        )

        cols = {}
        keywords = getattr(self, "_keywords", {})
        for line in self.process_lines(lines):
            # look for field[..]= ... column definitions
            cmatch = col_parser.match(line)
            if cmatch:
                name = cmatch.group("name")
                if len(name) > 24:
                    warn(
                        "The field name must be shorter than 24 characters.",
                        TdatFormatWarning,
                    )
                col = core.Column(name=name)
                col.meta = {}

                ctype = cmatch.group("ctype")
                if ctype in self._dtype_dict_in:
                    col.dtype = self._dtype_dict_in[ctype]
                elif "char" in ctype:
                    col.dtype = str
                else:
                    raise TdatFormatError(
                        f"Unrecognized or unsupported data type {ctype} for {col.name}."
                    )
                col.unit = cmatch.group("unit")
                col.format = cmatch.group("fmt")
                if len(str(ctype) + str(col.format)) > 23:
                    # max 24 characters with ":" separator
                    raise TdatFormatError(
                        "The type:fmt specifier has a\
                        maximum length of 24 characters. The offending line is:\
                            \n{line}\n"
                        + _STD_MSG
                    )
                col.description = f"{cmatch.group('desc')}".strip()
                for val in ["comment", "ucd", "index"]:
                    if cmatch.group(val) is not None:
                        text = cmatch.group(val).strip()
                        if (val == "comment") and (len(text) > 80):
                            warn(
                                TdatFormatWarning(
                                    "Comments are limited to 80 characters or less, truncating."
                                )
                            )
                        col.meta[val] = text[:80]
                cols[col.name] = col

        self.names = [
            val for line_field in self._line_fields.values() for val in line_field
        ]

        # check that cols and _line_fields are consistent or throw an error
        if len(self.names) != len(cols):
            colnames = [col.name for col in cols.values()]
            raise TdatFormatError(
                'The columns "field" descriptors are not consistent with '
                "the line[..] keyword.\n"
                f'"field" values: {colnames}\n'
                f"line[..] values: {self.names}"
            )
        self.cols = []
        for name in self.names:
            self.cols.append(cols[name])

    def write_comments(self, lines, meta):
        """Write comment lines in the header
        WRITE: Override the default write_comments to include <HEADER> as first line
        """
        lines.append("<HEADER>")
        if self.write_comment not in (False, None):
            for comment in meta.get("comments", []):
                lines.append(self.write_comment + comment.lstrip())

    def write(self, lines):
        """Write the Table out to a TDAT formatted file."""
        # the reader can handle alternate delimiters, but the writer should
        # comply with the current convention
        if self.splitter.delimiter.strip() != "|":
            warn(
                TdatFormatWarning(
                    "Delimiters other than the pipe character, '|', are deprecated. Using '|'."
                )
            )
        self.splitter.delimiter = "|"

        # Write the keywords and column descriptors
        keywords = deepcopy(self.table_meta.get("keywords", {}))
        meta_keys = ["keywords", "comments"]
        # In case a user puts a keyword directly in meta, instead of meta.keywords
        for key in [
            key.lower()
            for key in self.table_meta.keys()
            if key.lower() not in meta_keys
        ]:
            keywords[key] = self.table_meta.get(key)

        indices = [col.info.name for col in self.cols if col.info.indices != []]

        if "table_name" in keywords:
            if len(keywords["table_name"]) > 20:
                warn(
                    "'table_name' is too long, truncating to 20 characters",
                    TdatFormatWarning,
                )
            lines.append(f"table_name = {keywords.pop('table_name')[:20]}")
        else:
            warn(
                "'table_name' must be specified\n"  # noqa: ISC003
                + f"{_STD_MSG}\n"
                + "This should be specified in the Table.meta.\n"
                + "default value of 'astropy_table' being assigned.",
                TdatFormatWarning,
            )
            lines.append("table_name = astropy_table")

        # loop through optional table keywords
        table_keywords = [
            "table_description",
            "table_document_url",
            "table_security",
        ]
        table_keywords = [kw for kw in table_keywords if kw in keywords.keys()]
        for kw in ["record_delimiter", "field_delimiter"]:
            val = keywords.pop(kw, None)
            if val is not None:
                warn(TdatFormatWarning(f"Skipping deprecated keyword: {kw}."))

        for key in table_keywords:
            if key == "table_description":
                if len(keywords[key]) > 80:
                    warn(
                        "'table_description' is too long, truncating to 80 characters",
                        TdatFormatWarning,
                    )
                new_desc = keywords.pop(key)
                new_desc = new_desc[:80]
                lines.append(f"{key} = {new_desc}")
            elif key == "table_security":
                if keywords.get(key).lower() not in ("public", "private"):
                    warn(
                        "Value for table_security not recognized, should be public|private.",
                        TdatFormatWarning,
                    )
                lines.append(f"{key} = {keywords.pop(key)}")
            elif key in keywords:
                lines.append(f"{key} = {keywords.pop(key)}")

        # add table columns as fields
        lines.append("#")
        lines.append("# Table Parameters")
        lines.append("#")
        for col in self.cols:
            if str(col_type := col.info.dtype) in self._dtype_dict_out:
                ctype = self._dtype_dict_out[str(col_type)]
            elif col_type.kind == "i":
                ctype = "int4"
            elif col_type.kind == "f":
                ctype = "float8"
            elif col_type.kind == "U":
                ctype = f"char{col_type.itemsize // 4}"
            else:
                raise TdatFormatError(
                    f'Unrecognized data type `{col_type}` for column "{col.info.name}".'
                )
            col_name = col.info.name
            if len(col_name) >= 24:
                warn(
                    "The field name must be shorter than 24 characters, truncating.",
                    TdatFormatWarning,
                )
                col_name = col_name[:23]
            field_line = f"field[{col_name}] = {ctype}"

            col_info_meta = col.info.meta or {}
            if col.info.format is not None:
                field_line += f":{col.info.format}"
            if col.info.unit is not None:
                field_line += f"_{col.info.unit:cds}"
            if "ucd" in col_info_meta:
                field_line += f" [{col_info_meta['ucd']}]"
            if "index" in col_info_meta:
                field_line += f" ({col_info_meta['index']})"
            elif (indices != []) and (col.info.name == indices[0]):
                field_line += " (key)"
            elif col.info.name in indices:
                field_line += " (index)"
            if col.info.description is not None and col.info.description != "None":
                field_line += f" // {col.info.description}"
            elif "comment" in col_info_meta:
                field_line += " //"
            if "comment" in col_info_meta:
                field_line += f" // {col_info_meta['comment']}"
            lines.append(field_line)

        if len(keywords) != 0:
            if "parameter_defaults" in keywords:
                lines.append("#")
                lines.append(
                    f"{'parameter_defaults'} = {keywords.pop('parameter_defaults')}"
                )
        if len(keywords) != 0:
            lines.append("#")
            lines.append("# Virtual Parameters")
            lines.append("#")
            for key in keywords:
                lines.append(f"{key} = {keywords[key]}")
        lines.append("#")
        lines.append("# Data Format Specification")
        lines.append("#")
        lines.append(f"line[1] = {' '.join([col.info.name for col in self.cols])}")
        lines.append("#")


class TdatData(core.BaseData):
    """Data Reader for TDAT format"""

    comment = r"\s*(#|//)"
    write_comment = "# "
    splitter_class = TdatDataSplitter

    def get_data_lines(self, lines):
        """
        READ: Override the default get_data_lines to find start and end lines.
        """
        # Select lines between <DATA> and <END> in file
        try:
            start_line = (
                min(
                    i
                    for i, line in enumerate(lines)
                    if line.strip().upper() == "<DATA>"
                )
                + 1
            )
        except ValueError:
            raise TdatFormatError("<DATA> not found in file." + _STD_MSG)
        try:
            end_line = min(
                i for i, line in enumerate(lines) if line.strip().upper() == "<END>"
            )
        except ValueError:
            # <END> is an optional keyword to demarcate the end of the data.
            # If not present the end of the document is assumed to be the end
            # end of the data.
            end_line = len(lines) - 1

        self.data_lines = self.process_lines(lines[start_line:end_line])

    def str_vals(self):
        """WRITE: convert all values in table to a list of lists of strings.

        This sets the fill values and possibly column formats from the input
        formats={} keyword, then ends up calling table.pprint._pformat_col_iter()
        by a circuitous path. That function does the real work of formatting.
        Finally replace anything matching the fill_values.

        Returns
        -------
        values : list of list of str
        """
        self._set_fill_values(self.cols)
        self._set_col_formats()
        for col in self.cols:
            if np.issubdtype(col.dtype, np.integer):
                if np.any(col.data > 2**31 - 1) or np.any(col.data < -(2**31)):
                    raise TdatFormatError(
                        "Values cannot be converted to a TDAT compatible integer."
                    )
            col.str_vals = []
            for val in col.info.iter_str_vals():
                col.str_vals.append(val.replace("|", "\\|"))

        self._replace_vals(self.cols)
        return [col.str_vals for col in self.cols]

    def get_str_vals(self):
        """
        READ: Override the default get_str_vals to handle TDAT delimiters.
        """
        # if we have  delimiter from the header, use it.
        field_delimiter = getattr(self.header, "delimiter", self.splitter.delimiter)
        self.splitter.delimiter = field_delimiter
        record_delimiter = getattr(
            self.header, "record_delimiter", self.splitter.record_delimiter
        )
        allowed_delimiters = list(map(chr, range(1, 128)))
        if record_delimiter is not None:
            if record_delimiter not in allowed_delimiters:
                raise TdatFormatError(
                    f"Unsupported record delimiter: {record_delimiter}."
                )
            if record_delimiter == "|":
                raise TdatFormatError("'|' character reserved for field delimiter.")
        for c in field_delimiter:
            if c not in allowed_delimiters:
                raise TdatFormatError(f"Unsupported field delimiter: {c}.")
        nlines = len(self.header._line_fields)
        return self.splitter(self.data_lines, field_delimiter, record_delimiter, nlines)

    def write(self, lines):
        """
        WRITE: Override the default write to include <DATA> and <END> lines.
        """
        lines.append("<DATA>")
        super().write(lines)
        lines.append("<END>")


class TdatOutputter(core.TableOutputter):
    """
    Output the table as an astropy.table.Table object.
    """

    def __call__(self, cols, meta):
        """
        READ: Override the default outputter.
        TDAT files may (optionally) specify which field lines should be used as
        the primary index and secondary indices Astropy tables support adding
        indices after creation. This overwrite adds labeled indices on read.
        """
        # Sets col.data to numpy array and col.type to io.ascii Type class (e.g.
        # FloatType) for each col.
        self._convert_vals(cols)

        t_cols = [
            np.ma.MaskedArray(x.data, mask=x.mask)
            if hasattr(x, "mask") and np.any(x.mask)
            else x.data
            for x in cols
        ]
        out = core.Table(t_cols, names=[x.name for x in cols], meta=meta["table"])
        indices = []
        for col, out_col in zip(cols, out.columns.values()):
            for attr in ("format", "unit", "description"):
                if hasattr(col, attr):
                    setattr(out_col, attr, getattr(col, attr))
            if hasattr(col, "meta"):
                out_col.meta.update(col.meta)
                if "index" in col.meta:
                    if col.meta["index"] == "key":
                        indices.insert(0, col.name)
                    else:
                        indices.append(col.name)
        # Add indices, if specified
        if len(indices) > 0:
            for name in indices:
                out.add_index(name)
        return out


class Tdat(core.BaseReader):
    """TDAT format

    See: https://heasarc.gsfc.nasa.gov/docs/software/dbdocs/tdat.html

    Example::

      <HEADER>
      # # and // are comments
      table_name = example_table
      table_description = "Example table"
      #
      # Table Parameters
      #
      field[id] = integer [meta.id] (key) // Unique ID
      field[ra] = float:.4f_degree [pos.eq.ra] (index) // Right Ascension
      field[name] = char12 [meta.id] // Name
      #
      # Virtual Parameters
      #
      table_author = Example et al.
      #
      # Data Format Specification
      #
      line[1] = id name ra
      <DATA>
      1|TargetOne|1.0|
      2|TargetTwo|2.0|
      <END>

    The comments and keywords defined in the header, excepting common header
    section titles and blank comments, are available via the output table
    ``meta`` attribute::

      >>> from astropy.io import ascii
      >>> lines = ascii.tdat.make_example_data()
      >>> data = ascii.read(lines, format='tdat')
      >>> print(data.meta['comments'])
      ['# and // are comments']
      >>> for name, keyword in data.meta['keywords'].items():
      ...   print(name, keyword)
      table_name example_table
      table_description Example table
      table_author Example et al.

    When writing to the TDAT format, the header will be auto-populated by
    information in the Table, prioritizing information given in the Table.meta:

    **comments** : list or string, (optional)
      Table information which provide context. This information is
      included in the header preceding all other lines and commented
      out with #

    **keywords** : dict, (optional, recommended)
      Header keywords which will appear in the file as "name=value" lines.
      Of particular importance are table_name, table_description,
      and table_document_url.

    If there is no Table.meta, this writer will attempt to automatically
    generate the appropriate header information based on the table and
    column properties and the recommendations for the TDAT format by HEASARC.
    Column ``units`` are written using the CDS format.

    Example::

      >>> from astropy.table import Table
      >>> import sys
      >>> t = Table(names=('reference_id', 'RA', 'Name'),
      ...           data=[[1, 2, 3], [1.0, 2.0, 3.0], ['c', 'd', 'e']])
      >>> t.meta['table_name'] = "astropy_table"
      >>> t.write(sys.stdout, format="ascii.tdat")
      <HEADER>
      table_name = astropy_table
      #
      # Table Parameters
      #
      field[reference_id] = int4
      field[RA] = float8
      field[Name] = char1
      #
      # Data Format Specification
      #
      line[1] = reference_id RA Name
      #
      <DATA>
      1|1.0|c|
      2|2.0|d|
      3|3.0|e|
      <END>

    Including relevant metadata for the table and columns separately
    is possible with a mixture of attribute assignment and additions to the
    metadata::

      >>> from astropy.table import Table
      >>> from io import StringIO
      >>> t = Table(names=('reference_id', 'RA', 'Name'),
      ...           data=[[1, 2, 3], [1.0, 2.0, 3.0], ['c', 'd', 'e']])
      >>> t.meta["table_name"] = "example_table"
      >>> t.meta["table_description"] = "An example table for the tdat writer."
      >>> t.add_index('reference_id')
      >>> t.columns['reference_id'].meta['comment'] = "For internal reference only"
      >>> t.add_index('RA')
      >>> t.columns['RA'].unit = "degree"
      >>> t.columns['RA'].format = ".4f"
      >>> t.columns['RA'].meta['ucd'] = "pos.eq.ra"
      >>> t.columns['Name'].description = "The name of the source (if available)"
      >>> t.write(sys.stdout, format="ascii.tdat")
      <HEADER>
      table_name = example_table
      table_description = An example table for the tdat writer.
      #
      # Table Parameters
      #
      field[reference_id] = int4 (key) // // For internal reference only
      field[RA] = float8:.4f_deg [pos.eq.ra] (index)
      field[Name] = char1 // The name of the source (if available)
      #
      # Data Format Specification
      #
      line[1] = reference_id RA Name
      #
      <DATA>
      1|1.0000|c|
      2|2.0000|d|
      3|3.0000|e|
      <END>
    """

    _format_name = "tdat"
    _description = "HEASARC tdat format"
    _io_registry_can_write = True
    _io_registry_suffix = ".tdat"

    header_class = TdatHeader
    data_class = TdatData
    outputter_class = TdatOutputter



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from unittest import mock

import pytest

from astropy.io.fits import BinTableHDU, HDUList, Header, PrimaryHDU
from astropy.timeseries.io.kepler import kepler_fits_reader
from astropy.units import UnitsWarning
from astropy.utils.data import get_pkg_data_filename


def fake_header(extver, version, timesys, telescop):
    return Header(
        {
            "SIMPLE": "T",
            "BITPIX": 8,
            "NAXIS": 0,
            "EXTVER": extver,
            "VERSION": version,
            "TIMESYS": f"{timesys}",
            "TELESCOP": f"{telescop}",
        }
    )


def fake_hdulist(extver=1, version=2, timesys="TDB", telescop="KEPLER"):
    new_header = fake_header(extver, version, timesys, telescop)
    return [
        HDUList(
            hdus=[
                PrimaryHDU(header=new_header),
                BinTableHDU(header=new_header, name="LIGHTCURVE"),
            ]
        )
    ]


@mock.patch("astropy.io.fits.open", side_effect=fake_hdulist(telescop="MadeUp"))
def test_raise_telescop_wrong(mock_file):
    with pytest.raises(
        NotImplementedError,
        match=(
            r"MadeUp is not implemented, only KEPLER or TESS are supported through this"
            r" reader"
        ),
    ):
        kepler_fits_reader(None)


@mock.patch("astropy.io.fits.open", side_effect=fake_hdulist(extver=2))
def test_raise_extversion_kepler(mock_file):
    with pytest.raises(
        NotImplementedError, match=r"Support for KEPLER v2 files not yet implemented"
    ):
        kepler_fits_reader(None)


@mock.patch("astropy.io.fits.open", side_effect=fake_hdulist(extver=2, telescop="TESS"))
def test_raise_extversion_tess(mock_file):
    with pytest.raises(
        NotImplementedError, match=r"Support for TESS v2 files not yet implemented"
    ):
        kepler_fits_reader(None)


@mock.patch("astropy.io.fits.open", side_effect=fake_hdulist(timesys="TCB"))
def test_raise_timesys_kepler(mock_file):
    with pytest.raises(
        NotImplementedError,
        match=r"Support for TCB time scale not yet implemented in KEPLER reader",
    ):
        kepler_fits_reader(None)


@mock.patch(
    "astropy.io.fits.open", side_effect=fake_hdulist(timesys="TCB", telescop="TESS")
)
def test_raise_timesys_tess(mock_file):
    with pytest.raises(
        NotImplementedError,
        match=r"Support for TCB time scale not yet implemented in TESS reader",
    ):
        kepler_fits_reader(None)


@pytest.mark.remote_data(source="astropy")
def test_kepler_astropy():
    from astropy.units import UnitsWarning

    filename = get_pkg_data_filename("timeseries/kplr010666592-2009131110544_slc.fits")

    with pytest.warns(UnitsWarning):
        timeseries = kepler_fits_reader(filename)

    assert timeseries["time"].format == "isot"
    assert timeseries["time"].scale == "tdb"
    assert timeseries["sap_flux"].unit.to_string() == "electron / s"
    assert len(timeseries) == 14280
    assert len(timeseries.columns) == 20


@pytest.mark.remote_data(source="astropy")
def test_tess_astropy():
    filename = get_pkg_data_filename(
        "timeseries/hlsp_tess-data-alerts_tess_phot_00025155310-s01_tess_v1_lc.fits"
    )
    with pytest.warns((UserWarning, UnitsWarning)) as record:
        timeseries = kepler_fits_reader(filename)

    # we might hit some warnings more than once, but the exact sequence probably
    # does not matter too much, so we'll just try to match the *set* of unique warnings
    unique_warnings = {(wm.category, wm.message.args[0]) for wm in record}
    expected = {
        (UserWarning, "Ignoring 815 rows with NaN times"),
        (
            UnitsWarning,
            (
                "'BJD - 2457000, days' did not parse as fits unit: "
                "At col 0, Unit 'BJD' not supported by the FITS standard.  "
                "If this is meant to be a custom unit, define it with 'u.def_unit'. "
                "To have it recognized inside a file reader or other code, "
                "enable it with 'u.add_enabled_units'. "
                "For details, see https://docs.astropy.org/en/latest/units/combining_and_defining.html"
            ),
        ),
        (
            UnitsWarning,
            (
                "'pixels' did not parse as fits unit: "
                "At col 0, Unit 'pixels' not supported by the FITS standard. "
                "Did you mean pixel? If this is meant to be a custom unit, "
                "define it with 'u.def_unit'. To have it recognized inside a file "
                "reader or other code, enable it with 'u.add_enabled_units'. "
                "For details, see https://docs.astropy.org/en/latest/units/combining_and_defining.html"
            ),
        ),
        (
            UnitsWarning,
            (
                "'e-/s' did not parse as fits unit: "
                "At col 0, Unit 'e' not supported by the FITS standard.  "
                "If this is meant to be a custom unit, define it with 'u.def_unit'. "
                "To have it recognized inside a file reader or other code, "
                "enable it with 'u.add_enabled_units'. "
                "For details, see https://docs.astropy.org/en/latest/units/combining_and_defining.html"
            ),
        ),
    }
    assert unique_warnings == expected, (
        f"Got some unexpected warnings\n{unique_warnings - expected}"
    )
    assert timeseries["time"].format == "isot"
    assert timeseries["time"].scale == "tdb"
    assert timeseries["sap_flux"].unit.to_string() == "electron / s"
    assert len(timeseries) == 19261
    assert len(timeseries.columns) == 20



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Tests for putting velocity differentials into SkyCoord objects.

Note: the skyoffset velocity tests are in a different file, in
test_skyoffset_transformations.py
"""

import numpy as np
import pytest

from astropy import units as u
from astropy.coordinates import (
    ICRS,
    CartesianDifferential,
    CartesianRepresentation,
    Galactic,
    PrecessedGeocentric,
    RadialDifferential,
    SkyCoord,
    SphericalCosLatDifferential,
    SphericalDifferential,
    SphericalRepresentation,
    UnitSphericalCosLatDifferential,
    UnitSphericalDifferential,
    UnitSphericalRepresentation,
)
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.compat.optional_deps import HAS_SCIPY


def test_creation_frameobjs():
    i = ICRS(
        1 * u.deg, 2 * u.deg, pm_ra_cosdec=0.2 * u.mas / u.yr, pm_dec=0.1 * u.mas / u.yr
    )
    sc = SkyCoord(i)

    for attrnm in ["ra", "dec", "pm_ra_cosdec", "pm_dec"]:
        assert_quantity_allclose(getattr(i, attrnm), getattr(sc, attrnm))

    sc_nod = SkyCoord(ICRS(1 * u.deg, 2 * u.deg))

    for attrnm in ["ra", "dec"]:
        assert_quantity_allclose(getattr(sc, attrnm), getattr(sc_nod, attrnm))


def test_creation_attrs():
    sc1 = SkyCoord(
        1 * u.deg,
        2 * u.deg,
        pm_ra_cosdec=0.2 * u.mas / u.yr,
        pm_dec=0.1 * u.mas / u.yr,
        frame="fk5",
    )
    assert_quantity_allclose(sc1.ra, 1 * u.deg)
    assert_quantity_allclose(sc1.dec, 2 * u.deg)
    assert_quantity_allclose(sc1.pm_ra_cosdec, 0.2 * u.arcsec / u.kyr)
    assert_quantity_allclose(sc1.pm_dec, 0.1 * u.arcsec / u.kyr)

    sc2 = SkyCoord(
        1 * u.deg,
        2 * u.deg,
        pm_ra=0.2 * u.mas / u.yr,
        pm_dec=0.1 * u.mas / u.yr,
        differential_type=SphericalDifferential,
    )
    assert_quantity_allclose(sc2.ra, 1 * u.deg)
    assert_quantity_allclose(sc2.dec, 2 * u.deg)
    assert_quantity_allclose(sc2.pm_ra, 0.2 * u.arcsec / u.kyr)
    assert_quantity_allclose(sc2.pm_dec, 0.1 * u.arcsec / u.kyr)

    sc3 = SkyCoord(
        "1:2:3 4:5:6",
        pm_ra_cosdec=0.2 * u.mas / u.yr,
        pm_dec=0.1 * u.mas / u.yr,
        unit=(u.hour, u.deg),
    )

    assert_quantity_allclose(
        sc3.ra, 1 * u.hourangle + 2 * u.arcmin * 15 + 3 * u.arcsec * 15
    )
    assert_quantity_allclose(sc3.dec, 4 * u.deg + 5 * u.arcmin + 6 * u.arcsec)
    # might as well check with sillier units?
    assert_quantity_allclose(
        sc3.pm_ra_cosdec, 1.2776637006616473e-07 * u.arcmin / u.fortnight
    )
    assert_quantity_allclose(sc3.pm_dec, 6.388318503308237e-08 * u.arcmin / u.fortnight)


def test_creation_copy_basic():
    i = ICRS(
        1 * u.deg, 2 * u.deg, pm_ra_cosdec=0.2 * u.mas / u.yr, pm_dec=0.1 * u.mas / u.yr
    )
    sc = SkyCoord(i)
    sc_cpy = SkyCoord(sc)

    for attrnm in ["ra", "dec", "pm_ra_cosdec", "pm_dec"]:
        assert_quantity_allclose(getattr(sc, attrnm), getattr(sc_cpy, attrnm))


def test_creation_copy_rediff():
    sc = SkyCoord(
        1 * u.deg,
        2 * u.deg,
        pm_ra=0.2 * u.mas / u.yr,
        pm_dec=0.1 * u.mas / u.yr,
        differential_type=SphericalDifferential,
    )

    sc_cpy = SkyCoord(sc)
    for attrnm in ["ra", "dec", "pm_ra", "pm_dec"]:
        assert_quantity_allclose(getattr(sc, attrnm), getattr(sc_cpy, attrnm))

    sc_newdiff = SkyCoord(sc, differential_type=SphericalCosLatDifferential)
    reprepr = sc.represent_as(SphericalRepresentation, SphericalCosLatDifferential)
    assert_quantity_allclose(
        sc_newdiff.pm_ra_cosdec, reprepr.differentials["s"].d_lon_coslat
    )


def test_creation_cartesian():
    rep = CartesianRepresentation([10, 0.0, 0.0] * u.pc)
    dif = CartesianDifferential([0, 100, 0.0] * u.pc / u.Myr)
    rep = rep.with_differentials(dif)
    c = SkyCoord(rep)

    sdif = dif.represent_as(SphericalCosLatDifferential, rep)
    assert_quantity_allclose(c.pm_ra_cosdec, sdif.d_lon_coslat)


def test_useful_error_missing():
    sc_nod = SkyCoord(ICRS(1 * u.deg, 2 * u.deg))
    try:
        sc_nod.l
    except AttributeError as e:
        # this is double-checking the *normal* behavior
        msg_l = e.args[0]

    try:
        sc_nod.pm_dec
    except Exception as e:
        msg_pm_dec = e.args[0]

    assert "has no attribute" in msg_l
    assert "has no associated differentials" in msg_pm_dec


# ----------------------Operations on SkyCoords w/ velocities-------------------


# define some fixtures to get baseline coordinates to try operations with
@pytest.fixture(
    scope="module", params=[(False, False), (True, False), (False, True), (True, True)]
)
def sc(request):
    incldist, inclrv = request.param

    args = [1 * u.deg, 2 * u.deg]
    kwargs = {"pm_dec": 1 * u.mas / u.yr, "pm_ra_cosdec": 2 * u.mas / u.yr}
    if incldist:
        kwargs["distance"] = 213.4 * u.pc
    if inclrv:
        kwargs["radial_velocity"] = 61 * u.km / u.s

    return SkyCoord(*args, **kwargs)


@pytest.fixture(scope="module")
def scmany():
    return SkyCoord(
        ICRS(
            ra=[1] * 100 * u.deg,
            dec=[2] * 100 * u.deg,
            pm_ra_cosdec=np.random.randn(100) * u.mas / u.yr,
            pm_dec=np.random.randn(100) * u.mas / u.yr,
        )
    )


def test_accessors(sc, scmany):
    sc.data.differentials["s"]
    sph = sc.spherical
    gal = sc.galactic

    if sc.data.name.startswith("unit") and not sc.data.differentials[
        "s"
    ].name.startswith("unit"):
        # this xfail can be eliminated when issue #7028 is resolved
        pytest.xfail(".velocity fails if there is an RV but not distance")
    sc.velocity

    assert isinstance(sph, SphericalRepresentation)
    assert gal.data.differentials is not None

    scmany[0]
    sph = scmany.spherical
    gal = scmany.galactic

    assert isinstance(sph, SphericalRepresentation)
    assert gal.data.differentials is not None


def test_transforms(sc):
    trans = sc.transform_to("galactic")
    assert isinstance(trans.frame, Galactic)


def test_transforms_diff(sc):
    # note that arguably this *should* fail for the no-distance cases: 3D
    # information is necessary to truly solve this, hence the xfail
    if not sc.distance.unit.is_equivalent(u.m):
        pytest.xfail("Should fail for no-distance cases")
    else:
        trans = sc.transform_to(PrecessedGeocentric(equinox="B1975"))
        assert isinstance(trans.frame, PrecessedGeocentric)


@pytest.mark.skipif(not HAS_SCIPY, reason="Requires scipy")
def test_matching(sc, scmany):
    # just check that it works and yields something
    idx, d2d, d3d = sc.match_to_catalog_sky(scmany)


def test_constellations(sc):
    const = sc.get_constellation()
    assert const == "Pisces"


@pytest.mark.parametrize("sph_type", ["spherical", "unitspherical"])
def test_cartesian_to_spherical(sph_type):
    """Conversion to unitspherical should work, even if we lose distance."""
    c = SkyCoord(
        x=1 * u.kpc,
        y=0 * u.kpc,
        z=0 * u.kpc,
        v_x=10 * u.km / u.s,
        v_y=0 * u.km / u.s,
        v_z=4.74 * u.km / u.s,
        representation_type="cartesian",
    )
    c.representation_type = sph_type
    assert c.ra == 0
    assert c.dec == 0
    assert c.pm_ra == 0
    assert u.allclose(c.pm_dec, 1 * (u.mas / u.yr), rtol=1e-3)
    assert u.allclose(c.radial_velocity, 10 * (u.km / u.s))
    if sph_type == "spherical":
        assert u.allclose(c.distance, 1 * u.kpc)
    else:
        assert not hasattr(c, "distance")


@pytest.mark.parametrize(
    "diff_info, diff_cls",
    [
        ({"radial_velocity": [20, 30] * u.km / u.s}, RadialDifferential),
        (
            {
                "pm_ra": [2, 3] * u.mas / u.yr,
                "pm_dec": [-3, -4] * u.mas / u.yr,
                "differential_type": "unitspherical",
            },
            UnitSphericalDifferential,
        ),
        (
            {"pm_ra_cosdec": [2, 3] * u.mas / u.yr, "pm_dec": [-3, -4] * u.mas / u.yr},
            UnitSphericalCosLatDifferential,
        ),
    ],
    scope="class",
)
class TestDifferentialClassPropagation:
    """Test that going in between spherical and unit-spherical, we do not
    change differential type (since both can handle the same types).
    """

    def test_sc_unit_spherical_with_pm_or_rv_only(self, diff_info, diff_cls):
        sc = SkyCoord(ra=[10, 20] * u.deg, dec=[-10, 10] * u.deg, **diff_info)
        assert isinstance(sc.data, UnitSphericalRepresentation)
        assert isinstance(sc.data.differentials["s"], diff_cls)
        sr = sc.represent_as("spherical")
        assert isinstance(sr, SphericalRepresentation)
        assert isinstance(sr.differentials["s"], diff_cls)

    def test_sc_spherical_with_pm_or_rv_only(self, diff_info, diff_cls):
        sc = SkyCoord(
            ra=[10, 20] * u.deg,
            dec=[-10, 10] * u.deg,
            distance=1.0 * u.kpc,
            **diff_info,
        )
        assert isinstance(sc.data, SphericalRepresentation)
        assert isinstance(sc.data.differentials["s"], diff_cls)
        sr = sc.represent_as("unitspherical")
        assert isinstance(sr, UnitSphericalRepresentation)
        assert isinstance(sr.differentials["s"], diff_cls)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test the propagation of info on Quantity during operations."""

import copy

import numpy as np
import pytest

from astropy import units as u


def assert_info_equal(a, b, ignore=set()):
    a_info = a.info
    b_info = b.info
    for attr in (a_info.attr_names | b_info.attr_names) - ignore:
        if attr == "unit":
            assert a_info.unit.is_equivalent(b_info.unit)
        else:
            assert getattr(a_info, attr, None) == getattr(b_info, attr, None)


def assert_no_info(a):
    __tracebackhide__ = True
    assert "info" not in a.__dict__


class TestQuantityInfo:
    @classmethod
    def setup_class(cls):
        cls.q = u.Quantity(np.arange(1.0, 5.0), "m/s")
        cls.q.info.name = "v"
        cls.q.info.description = "air speed of a african swallow"

    def test_copy(self):
        q_copy1 = self.q.copy()
        assert_info_equal(q_copy1, self.q)
        q_copy2 = copy.copy(self.q)
        assert_info_equal(q_copy2, self.q)
        q_copy3 = copy.deepcopy(self.q)
        assert_info_equal(q_copy3, self.q)

    def test_slice(self):
        q_slice = self.q[1:3]
        assert_info_equal(q_slice, self.q)
        q_take = self.q.take([0, 1])
        assert_info_equal(q_take, self.q)

    def test_item(self):
        # Scalars do not get info set (like for Column); TODO: is this OK?
        q1 = self.q[1]
        assert_no_info(q1)
        q_item = self.q.item(1)
        assert_no_info(q_item)

    def test_iter(self):
        # Scalars do not get info set.
        for q in self.q:
            assert_no_info(q)
        for q in iter(self.q):
            assert_no_info(q)

    def test_change_to_equivalent_unit(self):
        q1 = self.q.to(u.km / u.hr)
        assert_info_equal(q1, self.q)
        q2 = self.q.si
        assert_info_equal(q2, self.q)
        q3 = self.q.cgs
        assert_info_equal(q3, self.q)
        q4 = self.q.decompose()
        assert_info_equal(q4, self.q)

    def test_reshape(self):
        q = self.q.reshape(-1, 1, 2)
        assert_info_equal(q, self.q)
        q2 = q.squeeze()
        assert_info_equal(q2, self.q)

    def test_insert(self):
        q = self.q.copy()
        q.insert(1, 1 * u.cm / u.hr)
        assert_info_equal(q, self.q)

    def test_unary_op(self):
        q = -self.q
        assert_no_info(q)

    def test_binary_op(self):
        q = self.q + self.q
        assert_no_info(q)

    def test_unit_change(self):
        q = self.q * u.s
        assert_no_info(q)
        q2 = u.s / self.q
        assert_no_info(q)

    def test_inplace_unit_change(self):
        # Not sure if it is logical to keep info here!
        q = self.q.copy()
        q *= u.s
        assert_info_equal(q, self.q, ignore={"unit"})

    def test_inplace_info_name_change(self):
        # see https://github.com/astropy/astropy/issues/17449
        q = self.q.copy()

        q.info.name = "test"
        assert q.info.name == "test"

        q.info.name = None
        assert q.info.name is None

        with pytest.raises(
            TypeError, match="Expected a str value, got 2.3 with type float"
        ):
            q.info.name = 2.3


class TestStructuredQuantity:
    @classmethod
    def setup_class(cls):
        value = np.array([(1.0, 2.0), (3.0, 4.0)], dtype=[("p", "f8"), ("v", "f8")])
        cls.q = u.Quantity(value, "m, m/s")
        cls.q.info.name = "pv"
        cls.q.info.description = "Location and speed"

    def test_keying(self):
        q_p = self.q["p"]
        assert_no_info(q_p)

    def test_slicing(self):
        q = self.q[:1]
        assert_info_equal(q, self.q)

    def test_item(self):
        # Scalars do not get info set.
        q = self.q[1]
        assert_no_info(q)


class TestQuantitySubclass:
    """Regression test for gh-14514: _new_view should __array_finalize__.

    But info should be propagated only for slicing, etc.
    """

    @classmethod
    def setup_class(cls):
        class MyQuantity(u.Quantity):
            def __array_finalize__(self, obj):
                super().__array_finalize__(obj)
                if hasattr(obj, "swallow"):
                    self.swallow = obj.swallow

        cls.my_q = MyQuantity([10.0, 20.0], u.m / u.s)
        cls.my_q.swallow = "African"
        cls.my_q_w_info = cls.my_q.copy()
        cls.my_q_w_info.info.name = "swallow"

    def test_setup(self):
        assert_no_info(self.my_q)
        assert self.my_q_w_info.swallow == self.my_q.swallow
        assert self.my_q_w_info.info.name == "swallow"

    def test_slice(self):
        slc1 = self.my_q[:1]
        assert slc1.swallow == self.my_q.swallow
        assert_no_info(slc1)
        slc2 = self.my_q_w_info[1:]
        assert slc2.swallow == self.my_q.swallow
        assert_info_equal(slc2, self.my_q_w_info)

    def test_op(self):
        square1 = self.my_q**2
        assert square1.swallow == self.my_q.swallow
        assert_no_info(square1)
        square2 = self.my_q_w_info**2
        assert square2.swallow == self.my_q.swallow
        assert_no_info(square2)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import warnings
from collections import defaultdict

import numpy as np
from matplotlib import rcParams
from matplotlib.artist import allow_rasterization
from matplotlib.text import Text

from astropy.utils.decorators import deprecated_renamed_argument
from astropy.utils.exceptions import AstropyDeprecationWarning

from .frame import RectangularFrame


def sort_using(X, Y):
    return [x for (y, x) in sorted(zip(Y, X))]


def _find_start_of_last_number(label):
    """
    Given a label, find the index of the start of the last numerical value
    in the label.
    """
    numerical_chars = "0123456789.+"
    if rcParams["axes.unicode_minus"] and not rcParams["text.usetex"]:
        numerical_chars += "\N{MINUS SIGN}"
    else:
        numerical_chars += "-"

    in_number = False
    for j in range(len(label) - 1, -1, -1):
        if in_number:
            if label[j] not in numerical_chars:
                return j + 1
        elif label[j] in numerical_chars:
            in_number = True


class TickLabels(Text):
    def __init__(self, frame, *args, **kwargs):
        self.clear()
        self._frame = frame
        super().__init__(*args, **kwargs)
        self.set_clip_on(True)
        self.set_visible_axes("all")
        self.set_pad(rcParams["xtick.major.pad"])
        self._exclude_overlapping = False
        self._simplify = True

        # Mapping from axis > list[bounding boxes]
        self._axis_bboxes = defaultdict(list)

        # Stale if either xy positions haven't been calculated, or if
        # something changes that requires recomputing the positions
        self._stale = True

        # Check rcParams
        if "color" not in kwargs:
            self.set_color(rcParams["xtick.color"])

        if "size" not in kwargs:
            self.set_size(rcParams["xtick.labelsize"])

    def clear(self):
        self.world = defaultdict(list)
        self.data = defaultdict(list)
        self.angle = defaultdict(list)
        self.text = defaultdict(list)
        self.disp = defaultdict(list)

    def add(
        self,
        axis=None,
        world=None,
        pixel=None,
        angle=None,
        text=None,
        axis_displacement=None,
        data=None,
    ):
        """
        Add a label.

        Parameters
        ----------
        axis : str
            Axis to add label to.
        world : Quantity
            Coordinate value along this axis.
        pixel : [float, float]
            Pixel coordinates of the label. Deprecated and no longer used.
        angle : float
            Angle of the label.
        text : str
            Label text.
        axis_displacement : float
            Displacement from axis.
        data : [float, float]
            Data coordinates of the label.
        """
        required_args = ["axis", "world", "angle", "text", "axis_displacement", "data"]
        if pixel is not None:
            warnings.warn(
                "Setting the pixel coordinates of a label does nothing and is"
                " deprecated, as these can only be accurately calculated when"
                " Matplotlib is drawing a figure. To prevent this warning pass the"
                f" following arguments as keyword arguments: {required_args}",
                AstropyDeprecationWarning,
            )
        if (
            axis is None
            or world is None
            or angle is None
            or text is None
            or axis_displacement is None
            or data is None
        ):
            raise TypeError(
                f"All of the following arguments must be provided: {required_args}"
            )

        self.world[axis].append(world)
        self.data[axis].append(data)
        self.angle[axis].append(angle)
        self.text[axis].append(text)
        self.disp[axis].append(axis_displacement)

        self._stale = True

    def sort(self):
        """
        Sort by axis displacement, which allows us to figure out which parts
        of labels to not repeat.
        """
        for axis in self.world:
            self.world[axis] = sort_using(self.world[axis], self.disp[axis])
            self.data[axis] = sort_using(self.data[axis], self.disp[axis])
            self.angle[axis] = sort_using(self.angle[axis], self.disp[axis])
            self.text[axis] = sort_using(self.text[axis], self.disp[axis])
            self.disp[axis] = sort_using(self.disp[axis], self.disp[axis])
        self._stale = True

    def simplify_labels(self):
        """
        Figure out which parts of labels can be dropped to avoid repetition.
        """
        self.sort()

        for axis in self.world:
            t1 = self.text[axis][0]
            for i in range(1, len(self.world[axis])):
                t2 = self.text[axis][i]

                if t1 == t2:
                    # In this case, we still need to preserve the last segment
                    # of the label. We search backwards from the end, and
                    # search for a number, and we then search for the first
                    # non-number (and non-decimal place) character we can find.
                    start = _find_start_of_last_number(t2)
                else:
                    for j in range(len(t1)):
                        if t1[j] != t2[j]:
                            start = _find_start_of_last_number(t2[: j + 1])
                            break

                if start != 0:
                    starts_dollar = t2.startswith("$")
                    self.text[axis][i] = t2[start:]
                    if starts_dollar:
                        self.text[axis][i] = "$" + self.text[axis][i]

                # Remove any empty LaTeX inline math mode string
                if self.text[axis][i] == "$$":
                    self.text[axis][i] = ""

                t1 = t2

        self._stale = True

    def set_pad(self, value):
        self._pad = value
        self._stale = True

    def get_pad(self):
        return self._pad

    def set_visible_axes(self, visible_axes):
        self._visible_axes = self._frame._validate_positions(visible_axes)
        self._stale = True

    def get_visible_axes(self):
        if self._visible_axes == "all":
            return list(self._frame.keys())
        else:
            return [x for x in self._visible_axes if x in self._frame or x == "#"]

    def set_exclude_overlapping(self, exclude_overlapping):
        self._exclude_overlapping = exclude_overlapping

    def set_simplify(self, simplify):
        self._simplify = simplify

    def _set_xy_alignments(self, renderer):
        """
        Compute and set the x, y positions and the horizontal/vertical alignment of
        each label.
        """
        if not self._stale:
            return

        if self._simplify:
            self.simplify_labels()
        text_size = renderer.points_to_pixels(self.get_size())

        visible_axes = self.get_visible_axes()
        self.xy = {axis: {} for axis in visible_axes}
        self.ha = {axis: {} for axis in visible_axes}
        self.va = {axis: {} for axis in visible_axes}

        for axis in visible_axes:
            for i in range(len(self.world[axis])):
                # In the event that the label is empty (which is not expected
                # but could happen in unforeseen corner cases), we should just
                # skip to the next label.
                if self.text[axis][i] == "":
                    continue

                x, y = self._frame.parent_axes.transData.transform(self.data[axis][i])
                pad = renderer.points_to_pixels(self.get_pad() + self._tick_out_size)

                if isinstance(self._frame, RectangularFrame):
                    # This is just to preserve the current results, but can be
                    # removed next time the reference images are re-generated.
                    if np.abs(self.angle[axis][i]) < 45.0:
                        ha = "right"
                        va = "bottom"
                        dx = -pad
                        dy = -text_size * 0.5
                    elif np.abs(self.angle[axis][i] - 90.0) < 45:
                        ha = "center"
                        va = "bottom"
                        dx = 0
                        dy = -text_size - pad
                    elif np.abs(self.angle[axis][i] - 180.0) < 45:
                        ha = "left"
                        va = "bottom"
                        dx = pad
                        dy = -text_size * 0.5
                    else:
                        ha = "center"
                        va = "bottom"
                        dx = 0
                        dy = pad

                    x = x + dx
                    y = y + dy

                else:
                    # This is the more general code for arbitrarily oriented
                    # axes

                    # Set initial position and find bounding box
                    self.set_text(self.text[axis][i])
                    self.set_position((x, y))
                    bb = super().get_window_extent(renderer)

                    # Find width and height, as well as angle at which we
                    # transition which side of the label we use to anchor the
                    # label.
                    width = bb.width
                    height = bb.height

                    # Project axis angle onto bounding box
                    ax = np.cos(np.radians(self.angle[axis][i]))
                    ay = np.sin(np.radians(self.angle[axis][i]))

                    # Set anchor point for label
                    if np.abs(self.angle[axis][i]) < 45.0:
                        dx = width
                        dy = ay * height
                    elif np.abs(self.angle[axis][i] - 90.0) < 45:
                        dx = ax * width
                        dy = height
                    elif np.abs(self.angle[axis][i] - 180.0) < 45:
                        dx = -width
                        dy = ay * height
                    else:
                        dx = ax * width
                        dy = -height

                    dx *= 0.5
                    dy *= 0.5

                    # Find normalized vector along axis normal, so as to be
                    # able to nudge the label away by a constant padding factor

                    dist = np.hypot(dx, dy)

                    ddx = dx / dist
                    ddy = dy / dist

                    dx += ddx * pad
                    dy += ddy * pad

                    x = x - dx
                    y = y - dy

                    ha = "center"
                    va = "center"

                self.xy[axis][i] = (x, y)
                self.ha[axis][i] = ha
                self.va[axis][i] = va

        self._stale = False

    def _get_bb(self, axis, i, renderer):
        """
        Get the bounding box of an individual label. n.b. _set_xy_alignment()
        must be called before this method.
        """
        if self.text[axis][i] == "":
            return

        self.set_text(self.text[axis][i])
        self.set_position(self.xy[axis][i])
        self.set_ha(self.ha[axis][i])
        self.set_va(self.va[axis][i])
        return super().get_window_extent(renderer)

    @property
    def _all_bboxes(self):
        # List of all tick label bounding boxes
        ret = []
        for axis in self._axis_bboxes:
            ret += self._axis_bboxes[axis]
        return ret

    def _set_existing_bboxes(self, bboxes):
        self._existing_bboxes = bboxes

    @allow_rasterization
    @deprecated_renamed_argument(old_name="bboxes", new_name=None, since="6.0")
    @deprecated_renamed_argument(old_name="ticklabels_bbox", new_name=None, since="6.0")
    @deprecated_renamed_argument(old_name="tick_out_size", new_name=None, since="6.0")
    def draw(self, renderer, bboxes=None, ticklabels_bbox=None, tick_out_size=None):
        # Reset bounding boxes
        self._axis_bboxes = defaultdict(list)

        if not self.get_visible():
            return

        self._set_xy_alignments(renderer)

        for axis in self.get_visible_axes():
            if axis == "#":
                continue

            for i in range(len(self.world[axis])):
                # This implicitly sets the label text, position, alignment
                bb = self._get_bb(axis, i, renderer)
                if bb is None:
                    continue

                # TODO: the problem here is that we might get rid of a label
                # that has a key starting bit such as -0:30 where the -0
                # might be dropped from all other labels.
                if (
                    not self._exclude_overlapping
                    or bb.count_overlaps(self._all_bboxes + self._existing_bboxes) == 0
                ):
                    super().draw(renderer)
                    self._axis_bboxes[axis].append(bb)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Table Indexing Implementation for Astropy Tables
================================================

This module implements indexing functionality for astropy Table objects, enabling fast
lookups and range queries on table columns. The indexing system uses various data
structures to maintain sorted representations of table data for efficient searching.

Note: these docs are intended for developers and are not rendered in the astropy
narrative docs.

Overview
--------
The astropy Table class supports creating indices on one or more columns to enable fast
data retrieval operations. When an index is created on a column (or combination of
columns), a sorted data structure is built that maps column values to their
corresponding row numbers in the original table.

The indices actually exist only on the columns. Calling ``Table.add_index()`` appends a
new `SlicedIndex` instance to the `col.indices` list for each of the indexed columns.
Adding an index always adds a `SlicedIndex` to the indices, with the slice initially
having ``start=0``, ``stop=len(col)-1`` and ``step=1``. The class has an ``original``
attribute which is initially ``True`` indicating that the index is not sliced.

The ``Table.indices`` property is a convenience method that returns a dynamically
created `TableIndices` object that collects the indices from table columns.
`TableIndices` is a list of `SlicedIndex` that also allows access by ``index_id``.

The ``Table.loc`` and ``Table.iloc`` properties return a new instance of
`TableLoc(table)` or `TableILoc(table)` each time.

Multiple indices can be defined for a single table. These indices are identified and
selected by an ``id`` tuple corresponding to the column names being indexed. The
`SlicedIndex` class defines an ``id`` property. In this module ``index_id`` is used for
arguments and variables referring to the index identifier.

Key Components
--------------

Core Classes:
    - `Index`: Wraps different engines and provides core index functions.
    - `SlicedIndex`: Contains an Index and handles original or sliced tables.
    - `TableIndices`: Container for all table indices, accessible by number and name.

Index Engines:
    - `SortedArray`: Array-based sorted container implementation (default)
    - `SCEngine`: SortedContainers-based implementation (when available)
    - `BST`: Binary search tree implementation for sorted data (not for production)

Query Interface Classes:
    - `TableLoc`: Value-based row retrieval using ``.loc[]``.
    - `TableLocIndices`: Return row indices (not table rows) for ``.loc_indices[]``
    - `TableILoc`: Return rows in based on sorted index order for ``.iloc[]``

Core Functionality
------------------
As a reminder of basic functionality.

Creating indices:
    Tables support adding indices via ``Table.add_index()``::

      t = Table({'a': [1, 3, 2, 5], 'b': ['x', 'y', 'z', 'w']})
      t.add_index('a')  # Single column index
      t.add_index(['a', 'b'])  # Multi-column composite index

Value-Based Lookups with TableLoc:
    The `TableLoc` class provides the `.loc[]` interface for value-based queries::

      t.loc[3]  # Find row where indexed column 'a' equals 3
      t.loc[2:4]  # Range query: rows where 'a' is between 2 and 4 (inclusive)
      t.loc[[1, 3, 5]]  # Multiple specific values
      t.loc.with_index('a')[3]  # Explicitly specify index column and value
      t.loc.with_index('a', 'b')[2, 'z']  # Multi-column composite key lookup

    Range queries support slice notation with inclusive bounds: - ``t.loc[2:5]`` returns
    rows where indexed values are between 2 and 5 - ``t.loc[:3]`` returns rows where
    values are ≤ 3 - ``t.loc[4:]`` returns rows where values are ≥ 4.

Value-Based Lookups with TableLocIndices:
    The `TableLocIndices` class provides the `.loc_indices[]` interface for value-based
    queries that return row indices instead of table rows. All of the examples for
    `t.loc` apply here as well, but with ``t.loc_indices`` returning indices instead::

      t.loc_indices[3]  # Find row index where indexed column 'a' equals 3

    This is useful when only row positions are needed.

Position-Based Access with TableILoc:
    The `TableILoc` class provides `.iloc[]` for accessing rows by their position in the
    sorted index order (not original table order)::

        t.iloc[0]  # First row in sorted order
        t.iloc[-1]  # Last row in sorted order
        t.iloc[1:3]  # Rows 1-2 in sorted order

Index Management:
    - `TableIndices`: Container class that allows retrieval of indices by index id
    - `get_index()`: Utility function to find existing indices on specified columns
    - `get_index_by_names()`: Find index by exact column name match

Sliced Table Handling
--------------------

When a table is sliced (e.g., `t_slice = t[2:8]`), the resulting table uses a
`SlicedIndex` that maintains a reference to the original table's index. This approach
avoids rebuilding indices for sliced views while ensuring lookup operations work
correctly.

The `SlicedIndex` class handles coordinate translation between the sliced view and the
original table:
    - Lookup operations translate slice-relative coordinates to original coordinates
    - Modification operations (add/remove rows) properly map coordinates
    - The `orig_coords()` method converts slice indices to original table indices
    - The `sliced_coords()` method converts original indices back to slice coordinates

The ``original`` attribute of the `SlicedIndex` instance gets set to `False` when the
parent table or column is sliced.

Query Processing Flow
--------------------

When a query is performed using `.loc[]`:

1. **Index Selection**: The system identifies which index to use based on the index id.
   If no id is specified, the primary key index is used.

2. **Query Type Detection**: The system determines the query type:
   - Single value: `t.loc[5]` → exact match lookup
   - Multiple values: `t.loc[[1, 3, 5]]` → multiple exact matches
   - Range query: `t.loc[2:6]` → range lookup with inclusive bounds

3. **Index Lookup**: The appropriate index engine performs the search:
   - Single/multiple lookups use `find()` method
   - Range queries use `range()` method with MinValue/MaxValue sentinels

4. **Coordinate Translation**: For sliced tables, row indices are translated from
   original coordinates to sliced coordinates using `sliced_coords()`

5. **Result Construction**: Row indices are used to construct the final Table slice or
   Row object returned to the user.

Data Structure Integration
-------------------------

The indexing system integrates with the Table's data management:
    - indices are automatically updated when rows are added or removed
    - Column value changes invalidate affected indices via the `replace()` method
    - Sliced tables maintain consistent index behavior through coordinate mapping
    - Memory usage is optimized by sharing index data between sliced views

TableLoc Implementation Details
------------------------------

The `TableLoc` class acts as a pseudo-list interface that translates value-based queries
into row-based table operations:

- **Query Parsing**: Handles various input formats (scalars, lists, slices, tuples)
- **Index Resolution**: Automatically selects the appropriate index or uses primary key
- **Bounds Handling**: Converts Python slice notation to inclusive range queries
- **Result Formatting**: Returns single Row objects for scalar queries, Table slices
  otherwise

The `TableLocIndices` variant returns raw row indices instead of table data, useful for
advanced indexing operations or when only row positions are needed.

Implementation Notes
-------------------

The index system is designed to be engine-agnostic, with different backend
implementations (BST, SortedArray, etc.) providing the same interface. This allows for
performance optimization based on data characteristics and available dependencies.

The TableLoc classes use a unified internal method `_get_row_idxs_as_list_or_int()` to
handle the complexity of different query types while maintaining consistent behavior
across the `.loc[]`, `.iloc[]`, and `.loc_indices[]` interfaces.

Protocol for engine classes
---------------------------
The `Index` class can use several implementations as its engine. Any implementation must
implement the following:

- __init__(data, row_index) : initialize index based on key/row list pairs
- add(key, row) -> None : add (key, row) to existing data
- remove(key, data=None) -> boolean : remove data from self[key], or all of
    self[key] if data is None
- shift_left(row) -> None : decrement row numbers after row
- shift_right(row) -> None : increase row numbers >= row
- find(key) -> list : list of rows corresponding to key
- range(lower, upper, bounds) -> list : rows in self[k] where k is between
    lower and upper (<= or < based on bounds)
- sort() -> None : make row order align with key order
- sorted_data() -> list of rows in sorted order (by key)
- replace_rows(row_map) -> None : replace row numbers based on slice
- items() -> list of tuples of the form (key, data)

Copying
-------
When a Table is initialized from another Table, indices are deep-copied and their
columns are set to the columns of the new Table.

Column creation: - Column(c) -> deep copy of indices - c[[1, 2]] -> deep copy and
reordering of indices - c[1:2] -> reference - array.view(Column) -> no indices
"""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Protocol, runtime_checkable

import numpy as np

from astropy.utils.decorators import deprecated

from .sorted_array import SortedArray

if TYPE_CHECKING:
    from collections.abc import Hashable, Mapping, Sequence
    from numbers import Integral

    from . import QTable, Table


@runtime_checkable
class IndexEngine(Protocol):
    """Protocol defining an index engine class"""

    # Taken from soco.py
    def __init__(self, data, row_index, unique=False): ...
    # using *args and **kwargs here so existing implementations,
    # while slightly inconsistent with one another, remain conform without change.
    def add(self, key: tuple, *args: int | None, **kwargs: int | None) -> None: ...
    # if we manage to remove the one implementation that's inconsistent with it,
    # the actual signature should be:
    # def add(self, key: tuple, row: int) -> None: ...
    def find(self, key: tuple) -> Sequence[Integral]: ...
    def remove(self, key: tuple, data: int) -> bool: ...
    def shift_left(self, row: int) -> None: ...
    def shift_right(self, row: int) -> None: ...
    def items(self) -> list[tuple[Hashable, list[Integral]]]: ...
    def sort(self) -> None: ...
    def sorted_data(self) -> None: ...
    def range(
        self,
        lower: tuple[Hashable, ...] | None,
        upper: tuple[Hashable, ...] | None,
        bounds: tuple[bool, bool],
    ) -> list[int]: ...
    def replace_rows(self, row_map: Mapping[int, int]) -> None: ...


class QueryError(ValueError):
    """
    Indicates that a given index cannot handle the supplied query.
    """


class Index:
    """
    The Index class makes it possible to maintain indices
    on columns of a Table, so that column values can be queried
    quickly and efficiently. Column values are stored in lexicographic
    sorted order, which allows for binary searching in O(log n).

    Parameters
    ----------
    columns : list or None
        List of columns on which to create an index. If None,
        create an empty index for purposes of deep copying.
    engine : type, instance, or None
        Indexing engine class to use (from among SortedArray, BST,
        and SCEngine) or actual engine instance.
        If the supplied argument is None (by default), use SortedArray.
    unique : bool (defaults to False)
        Whether the values of the index must be unique
    """

    def __init__(self, columns, engine=None, unique=False):
        # Local imports to avoid import problems.
        from astropy.time import Time

        from .table import Column

        if columns is not None:
            columns = list(columns)

        # by default, use SortedArray
        if engine is None:
            engine = SortedArray

        # Validate engine. This catches an easy mistake of `t.add_index("a", "b")`.
        engine_cls = engine if isinstance(engine, type) else engine.__class__
        if not issubclass(engine_cls, IndexEngine):
            raise TypeError(
                f"engine must be an Engine class or instance, got {engine!r} instead."
            )

        if engine is not engine_cls:
            # create from data
            self.engine = engine_cls
            self.data = engine
            self.columns = columns
            return

        self.engine = engine

        if columns is None:  # this creates a special exception for deep copying
            columns = []
            data = []
            row_index = []
        elif len(columns) == 0:
            raise ValueError("Cannot create index without at least one column")

        self.columns = columns

        if len(columns) == 1:
            col = columns[0]
            row_index = Column(col.argsort(kind="stable"))
            data = self.table_cls([col[row_index]])
        elif len(columns) > 1:
            num_rows = len(columns[0])

            # replace Time columns with approximate form and remainder
            new_columns = []
            for col in columns:
                if isinstance(col, Time):
                    new_columns.append(col.jd)
                    remainder = col - col.__class__(
                        col.jd, format="jd", scale=col.scale
                    )
                    new_columns.append(remainder.jd)
                else:
                    new_columns.append(col)

            # sort the table lexicographically and keep row numbers
            table = self.table_cls(columns + [np.arange(num_rows)], copy_indices=False)
            sort_columns = new_columns[::-1]
            try:
                lines = table[np.lexsort(sort_columns)]
            except TypeError:  # arbitrary mixins might not work with lexsort
                lines = table[table.argsort(kind="stable")]
            data = lines[lines.colnames[:-1]]
            row_index = lines[lines.colnames[-1]]

        self.data = self.engine(data, row_index, unique=unique)

    def __len__(self):
        """
        Number of rows in index.
        """
        return len(self.columns[0])

    @property
    def table_cls(self) -> type[Table | QTable]:
        """
        The Table class (Table or QTable) corresponding to the
        type of columns in this index.
        """
        from astropy.table import QTable, Table
        from astropy.units import Quantity

        has_quantity = any(isinstance(col, Quantity) for col in self.columns)
        return QTable if has_quantity else Table

    def replace_col(self, prev_col, new_col):
        """
        Replace an indexed column with an updated reference.

        Parameters
        ----------
        prev_col : Column
            Column reference to replace
        new_col : Column
            New column reference
        """
        self.columns[self.col_position(prev_col.info.name)] = new_col

    def reload(self):
        """
        Recreate the index based on data in self.columns.
        """
        self.__init__(self.columns, engine=self.engine)

    def col_position(self, col_name):
        """
        Return the position of col_name in self.columns.

        Parameters
        ----------
        col_name : str
            Name of column to look up
        """
        for i, c in enumerate(self.columns):
            if c.info.name == col_name:
                return i
        raise ValueError(f"Column does not belong to index: {col_name}")

    def insert_row(self, pos, vals, columns):
        """
        Insert a new row from the given values.

        Parameters
        ----------
        pos : int
            Position at which to insert row
        vals : list or tuple
            List of values to insert into a new row
        columns : list
            Table column references
        """
        key = [None] * len(self.columns)
        for i, col in enumerate(columns):
            try:
                key[self.col_position(col.info.name)] = vals[i]
            except ValueError:  # not a member of index
                continue
        num_rows = len(self.columns[0])
        if pos < num_rows:
            # shift all rows >= pos to the right
            self.data.shift_right(pos)
        self.data.add(tuple(key), pos)

    def get_row_specifier(self, row_specifier):
        """
        Return an iterable corresponding to the
        input row specifier.

        Parameters
        ----------
        row_specifier : int, list, ndarray, or slice
        """
        if isinstance(row_specifier, (int, np.integer)):
            # single row
            return (row_specifier,)
        elif isinstance(row_specifier, (list, np.ndarray)):
            return row_specifier
        elif isinstance(row_specifier, slice):
            col_len = len(self.columns[0])
            return range(*row_specifier.indices(col_len))
        raise ValueError(
            f"Expected int, array of ints, or slice but got {row_specifier} "
            "in remove_rows"
        )

    def remove_rows(self, row_specifier):
        """
        Remove the given rows from the index.

        Parameters
        ----------
        row_specifier : int, list, ndarray, or slice
            Indicates which row(s) to remove
        """
        rows = []

        # To maintain the correct row order, we loop twice,
        # deleting rows first and then reordering the remaining rows
        for row in self.get_row_specifier(row_specifier):
            self.remove_row(row, reorder=False)
            rows.append(row)
        # second pass - row order is reversed to maintain
        # correct row numbers
        for row in sorted(rows, reverse=True):
            self.data.shift_left(row)

    def remove_row(self, row, reorder=True):
        """
        Remove the given row from the index.

        Parameters
        ----------
        row : int
            Table row number to remove in original table order
        reorder : bool
            Whether to reorder indices after removal
        """
        # For removal, form a key consisting of column values in this row, where
        # self.columns references the original columns.
        if not self.data.remove(tuple(col[row] for col in self.columns), row):
            raise ValueError(f"Could not remove row {row} from index")
        # decrement the row number of all later rows
        if reorder:
            self.data.shift_left(row)

    def find(self, key):
        """
        Return the row values corresponding to key, in sorted order.

        Parameters
        ----------
        key : tuple
            Values to search for in each column
        """
        return self.data.find(key)

    @deprecated(since="7.2.0")
    def same_prefix(self, key):
        """
        Return rows whose keys contain the supplied key as a prefix.

        Parameters
        ----------
        key : tuple
            Prefix for which to search
        """
        return self.same_prefix_range(key, key, (True, True))

    @deprecated(since="7.2.0")
    def same_prefix_range(self, lower, upper, bounds=(True, True)):
        """
        Return rows whose keys have a prefix in the given range.

        Parameters
        ----------
        lower : tuple
            Lower prefix bound
        upper : tuple
            Upper prefix bound
        bounds : tuple (x, y) of bools
            Indicates whether the search should be inclusive or
            exclusive with respect to the endpoints. The first
            argument x corresponds to an inclusive lower bound,
            and the second argument y to an inclusive upper bound.
        """
        from .bst import MaxValue, MinValue

        n = len(lower)
        ncols = len(self.columns)
        a = MinValue() if bounds[0] else MaxValue()
        b = MaxValue() if bounds[1] else MinValue()
        # [x, y] search corresponds to [(x, min), (y, max)]
        # (x, y) search corresponds to ((x, max), (x, min))
        lower = lower + tuple((ncols - n) * [a])
        upper = upper + tuple((ncols - n) * [b])
        return self.data.range(lower, upper, bounds)

    def range(
        self,
        lower: tuple[Hashable, ...] | None,
        upper: tuple[Hashable, ...] | None,
        bounds: tuple[bool, bool] = (True, True),
    ) -> list[int]:
        """
        Return rows within the given range.

        Parameters
        ----------
        lower : tuple, None
            Lower search bound (no lower bound if None)
        upper : tuple, None
            Upper search bound (no upper bound if None)
        bounds : tuple (x, y) of bools
            Indicates whether the search should be inclusive or
            exclusive with respect to the endpoints. The first
            argument x corresponds to an inclusive lower bound,
            and the second argument y to an inclusive upper bound.
        """
        return self.data.range(lower, upper, bounds)

    def replace(self, row, col_name, val):
        """
        Replace the value of a column at a given position.

        Parameters
        ----------
        row : int
            Row number to modify
        col_name : str
            Name of the Column to modify
        val : col.info.dtype
            Value to insert at specified row of col
        """
        self.remove_row(row, reorder=False)
        key = [c[row] for c in self.columns]
        key[self.col_position(col_name)] = val
        self.data.add(tuple(key), row)

    def replace_rows(self, col_slice):
        """
        Modify rows in this index to agree with the specified
        slice. For example, given an index
        {'5': 1, '2': 0, '3': 2} on a column ['2', '5', '3'],
        an input col_slice of [2, 0] will result in the relabeling
        {'3': 0, '2': 1} on the sliced column ['3', '2'].

        Parameters
        ----------
        col_slice : list
            Indices to slice
        """
        row_map = {row: i for i, row in enumerate(col_slice)}
        self.data.replace_rows(row_map)

    def sort(self):
        """
        Make row numbers follow the same sort order as the keys
        of the index.
        """
        self.data.sort()

    def sorted_data(self):
        """
        Returns a list of rows in sorted order based on keys;
        essentially acts as an argsort() on columns.
        """
        return self.data.sorted_data()

    def __getitem__(self, item):
        """
        Returns a sliced version of this index.

        Parameters
        ----------
        item : slice
            Input slice

        Returns
        -------
        SlicedIndex
            A sliced reference to this index.
        """
        return SlicedIndex(self, item)

    def __repr__(self):
        col_names = tuple(col.info.name for col in self.columns)
        return f"<{self.__class__.__name__} columns={col_names} data={self.data}>"

    def __deepcopy__(self, memo):
        """
        Return a deep copy of this index.

        Notes
        -----
        The default deep copy must be overridden to perform
        a shallow copy of the index columns, avoiding infinite recursion.

        Parameters
        ----------
        memo : dict
        """
        # Bypass Index.__new__ to create an actual Index, not a SlicedIndex.
        index = super().__new__(self.__class__)
        index.__init__(None, engine=self.engine)
        index.data = deepcopy(self.data, memo)
        index.columns = self.columns[:]  # new list, same columns
        memo[id(self)] = index
        return index


class SlicedIndex:
    """Class that provides the column-level interface for index operations.

    This is wrapper around an Index object that allows for efficient slicing
    functionality, but it is also used for an unsliced index. Since numpy expects array
    slices to provide an actual data view, a SlicedIndex retrieves data directly from
    the original index and then adapt it to the sliced coordinate system as appropriate.

    Parameters
    ----------
    index : Index
        The original Index reference
    index_slice : tuple, slice
        The slice to which this SlicedIndex corresponds
    original : bool
        Whether this SlicedIndex represents the original index itself. For the most part
        this is similar to index[:] but certain copying operations are avoided, and the
        slice retains the length of the actual index despite modification.
    """

    def __init__(self, index: Index, index_slice, original=False):
        self.index = index
        self.original = original
        self._frozen = False

        if isinstance(index_slice, tuple):
            self.start, self._stop, self.step = index_slice
        elif isinstance(index_slice, slice):  # index_slice is an actual slice
            num_rows = len(index.columns[0])
            self.start, self._stop, self.step = index_slice.indices(num_rows)
        else:
            raise TypeError("index_slice must be tuple or slice")

    @property
    def id(self) -> tuple[str, ...]:
        """Identifier for this index as tuple of index column names"""
        return tuple(col.info.name for col in self.columns)

    @property
    def length(self):
        return 1 + (self.stop - self.start - 1) // self.step

    @property
    def stop(self):
        """
        The stopping position of the slice, or the end of the
        index if this is an original slice.
        """
        return len(self.index) if self.original else self._stop

    def __getitem__(self, item):
        """
        Returns another slice of this Index slice.

        Parameters
        ----------
        item : slice
            Index slice
        """
        if self.length <= 0:
            # empty slice
            return SlicedIndex(self.index, slice(1, 0))
        start, stop, step = item.indices(self.length)
        new_start = self.orig_coords(start)
        new_stop = self.orig_coords(stop)
        new_step = self.step * step
        return SlicedIndex(self.index, (new_start, new_stop, new_step))

    def sliced_coords(self, rows):
        """
        Convert the input rows to the sliced coordinate system.

        Parameters
        ----------
        rows : list
            Rows in the original coordinate system

        Returns
        -------
        sliced_rows : list
            Rows in the sliced coordinate system
        """
        if self.original:
            return rows
        else:
            rows = np.array(rows)
            row0 = rows - self.start
            if self.step != 1:
                correct_mod = np.mod(row0, self.step) == 0
                row0 = row0[correct_mod]
            if self.step > 0:
                ok = (row0 >= 0) & (row0 < self.stop - self.start)
            else:
                ok = (row0 <= 0) & (row0 > self.stop - self.start)
            return row0[ok] // self.step

    def orig_coords(self, row):
        """
        Convert the input row from sliced coordinates back
        to original coordinates.

        Parameters
        ----------
        row : int
            Row in the sliced coordinate system

        Returns
        -------
        orig_row : int
            Row in the original coordinate system
        """
        return row if self.original else self.start + row * self.step

    def find(self, key):
        return self.sliced_coords(self.index.find(key))

    def where(self, col_map):
        return self.sliced_coords(self.index.where(col_map))

    def range(self, lower, upper):
        return self.sliced_coords(self.index.range(lower, upper))

    def same_prefix(self, key):
        return self.sliced_coords(self.index.same_prefix(key))

    def sorted_data(self):
        return self.sliced_coords(self.index.sorted_data())

    def replace(self, row, col, val):
        if not self._frozen:
            self.index.replace(self.orig_coords(row), col, val)

    def get_index_or_copy(self):
        if not self.original:
            # replace self.index with a new object reference
            self.index = deepcopy(self.index)
        return self.index

    def insert_row(self, pos, vals, columns):
        if not self._frozen:
            self.get_index_or_copy().insert_row(self.orig_coords(pos), vals, columns)

    def get_row_specifier(self, row_specifier):
        return [
            self.orig_coords(x) for x in self.index.get_row_specifier(row_specifier)
        ]

    def remove_rows(self, row_specifier):
        if not self._frozen:
            self.get_index_or_copy().remove_rows(row_specifier)

    def replace_rows(self, col_slice):
        if not self._frozen:
            self.index.replace_rows([self.orig_coords(x) for x in col_slice])

    def sort(self):
        if not self._frozen:
            self.get_index_or_copy().sort()

    def __repr__(self):
        slice_str = (
            "" if self.original else f" slice={self.start}:{self.stop}:{self.step}"
        )
        return (
            f"<{self.__class__.__name__} original={self.original}{slice_str}"
            f" index={self.index}>"
        )

    def replace_col(self, prev_col, new_col):
        self.index.replace_col(prev_col, new_col)

    def reload(self):
        self.index.reload()

    def col_position(self, col_name):
        return self.index.col_position(col_name)

    def get_slice(self, col_slice, item):
        """
        Return a newly created index from the given slice.

        Parameters
        ----------
        col_slice : Column object
            Already existing slice of a single column
        item : list or ndarray
            Slice for retrieval
        """
        if len(self.columns) == 1:
            index = Index([col_slice], engine=self.data.__class__)
            return self.__class__(index, slice(0, 0, None), original=True)

        t = self.index.table_cls(self.columns, copy_indices=False)
        with t.index_mode("discard_on_copy"):
            new_cols = t[item].columns.values()
        index = Index(new_cols, engine=self.data.__class__)
        return self.__class__(index, slice(0, 0, None), original=True)

    @property
    def columns(self):
        return self.index.columns

    @property
    def data(self):
        return self.index.data


def get_index(table, table_copy=None, names=None):
    """
    Inputs a table and some subset of its columns as table_copy.
    List or tuple containing names of columns as names,and returns an index
    corresponding to this subset or list or None if no such index exists.

    Parameters
    ----------
    table : `Table`
        Input table
    table_copy : `Table`, optional
        Subset of the columns in the ``table`` argument
    names : list, tuple, optional
        Subset of column names in the ``table`` argument

    Returns
    -------
    Index of columns or None

    """
    if names is not None and table_copy is not None:
        raise ValueError(
            'one and only one argument from "table_copy" or "names" is required'
        )

    if names is None and table_copy is None:
        raise ValueError(
            'one and only one argument from "table_copy" or "names" is required'
        )

    if names is not None:
        names = set(names)
    else:
        names = set(table_copy.colnames)

    if not names <= set(table.colnames):
        raise ValueError(f"{names} is not a subset of table columns")

    for name in names:
        for index in table[name].info.indices:
            if {col.info.name for col in index.columns} == names:
                return index

    return None


def get_index_by_names(table: Table, names: tuple | list) -> SlicedIndex | None:
    """
    Return index in ``table`` keyed by ``names``, or `None` if not found.

    Parameters
    ----------
    table : `Table`
        Input table
    names : tuple, list
        Key names of table index to find

    Returns
    -------
    SlicedIndex or None
        Table Index or None
    """
    index_id = tuple(names)
    for index in table.indices:
        if index_id == index.id:
            return index
    return None


class _IndexModeContext:
    """
    A context manager that allows for special indexing modes, which
    are intended to improve performance. Currently the allowed modes
    are "freeze", in which indices are not modified upon column modification,
    "copy_on_getitem", in which indices are copied upon column slicing,
    and "discard_on_copy", in which indices are discarded upon table
    copying/slicing.
    """

    _col_subclasses = {}

    def __init__(self, table, mode):
        """
        Parameters
        ----------
        table : Table
            The table to which the mode should be applied
        mode : str
            Either 'freeze', 'copy_on_getitem', or 'discard_on_copy'.
            In 'discard_on_copy' mode,
            indices are not copied whenever columns or tables are copied.
            In 'freeze' mode, indices are not modified whenever columns are
            modified; at the exit of the context, indices refresh themselves
            based on column values. This mode is intended for scenarios in
            which one intends to make many additions or modifications on an
            indexed column.
            In 'copy_on_getitem' mode, indices are copied when taking column
            slices as well as table slices, so col[i0:i1] will preserve
            indices.
        """
        self.table = table
        self.mode = mode
        # Used by copy_on_getitem
        self._orig_classes = []
        if mode not in ("freeze", "discard_on_copy", "copy_on_getitem"):
            raise ValueError(
                "Expected a mode of either 'freeze', "
                "'discard_on_copy', or 'copy_on_getitem', got "
                f"'{mode}'"
            )

    def __enter__(self):
        if self.mode == "discard_on_copy":
            self.table._copy_indices = False
        elif self.mode == "copy_on_getitem":
            for col in self.table.columns.values():
                self._orig_classes.append(col.__class__)
                col.__class__ = self._get_copy_on_getitem_shim(col.__class__)
        else:
            for index in self.table.indices:
                index._frozen = True

    def __exit__(self, exc_type, exc_value, traceback):
        if self.mode == "discard_on_copy":
            self.table._copy_indices = True
        elif self.mode == "copy_on_getitem":
            for col in reversed(self.table.columns.values()):
                col.__class__ = self._orig_classes.pop()
        else:
            for index in self.table.indices:
                index._frozen = False
                index.reload()

    def _get_copy_on_getitem_shim(self, cls):
        """
        This creates a subclass of the column's class which overrides that
        class's ``__getitem__``, such that when returning a slice of the
        column, the relevant indices are also copied over to the slice.

        Ideally, rather than shimming in a new ``__class__`` we would be able
        to just flip a flag that is checked by the base class's
        ``__getitem__``.  Unfortunately, since the flag needs to be a Python
        variable, this slows down ``__getitem__`` too much in the more common
        case where a copy of the indices is not needed.  See the docstring for
        ``astropy.table._column_mixins`` for more information on that.
        """
        if cls in self._col_subclasses:
            return self._col_subclasses[cls]

        def __getitem__(self, item):
            value = cls.__getitem__(self, item)
            if type(value) is type(self):
                value = self.info.slice_indices(value, item, len(self))

            return value

        clsname = f"_{cls.__name__}WithIndexCopy"

        new_cls = type(str(clsname), (cls,), {"__getitem__": __getitem__})

        self._col_subclasses[cls] = new_cls

        return new_cls


class TableIndices(list):
    """
    List-subclass of table indices allowing for retrieval by column name(s).

    Parameters
    ----------
    lst : list
        List of indices
    """

    def __init__(self, lst):
        super().__init__(lst)

    def __getitem__(self, item) -> Index:
        """
        Retrieve an item from the list of indices.

        Parameters
        ----------
        item : int, str, tuple, or list
            Position in list or name(s) of indexed column(s)
        """
        if isinstance(item, str):
            item = [item]
        if isinstance(item, (list, tuple)):
            item = list(item)
            for index in self:
                try:
                    for name in item:
                        index.col_position(name)
                    if len(index.columns) == len(item):
                        return index
                except ValueError:
                    pass
            # index search failed
            raise IndexError(f"No index found for {item}")

        return super().__getitem__(item)


@deprecated(
    since="7.2.0",
    message="""\
Calling `Table.loc/iloc/loc_indices[index_id, item]` to select `item` from index
`index_id` is deprecated. Instead select the index using the syntax
`Table.loc/iloc/loc_indices.with_index(index_id)[item]`.
""",
)
def interpret_item_as_index_id_and_item(item: tuple) -> tuple:
    """Interpret the item as a (index_id, item) tuple."""
    index_id, item = item
    return index_id, item


class TableLoc:
    """
    Pseudo-list of Table rows allowing for retrieval of rows by indexed column values.

    Parameters
    ----------
    table : Table
        Indexed table to use
    index_id : tuple or None
        If not None, the index id as a tuple to use for all retrievals. If None
        (default), the primary key index is used.
    """

    def __init__(self, table: Table, index_id: tuple[str, ...] | None = None):
        self.table = table
        self.indices = table.indices
        self.index_id = index_id

    def __repr__(self):
        index_id = self.index_id
        if index_id is None:
            index_id = self.table.primary_key
        id_repr = index_id[0] if len(index_id) == 1 else index_id

        return (
            f"<{self.__class__.__name__} index_id={repr(id_repr)} "
            f"id(table)={id(self.table)}>"
        )

    def with_index(self, *index_id):
        """Return a new instance of this class for ``index_id``

        Parameters
        ----------
        index_id : str, tuple[str, ...], or list[str]
            Identifier of the index to use

        Examples
        --------
        >>> from astropy.table import QTable
        >>> t = QTable({'a': [1, 2, 3], 'b': [4, 5, 6], 'c': [7, 8, 9]})
        >>> t.add_index('a')
        >>> t.add_index(['b', 'c'])
        >>> t.loc.with_index('a')[2]  # doctest: +IGNORE_OUTPUT
        >>> t.loc.with_index('b', 'c')[5, 8]  # doctest: +IGNORE_OUTPUT
        >>> t.loc.with_index(['b', 'c'])[5, 8]  # doctest: +IGNORE_OUTPUT
        """
        if len(index_id) == 1 and isinstance(index_id[0], (tuple, list)):
            index_id = tuple(index_id[0])
        return self.__class__(self.table, index_id)

    def _get_index_id_and_item(self, item):
        index_id = self.index_id
        if self.index_id is None:
            if isinstance(item, tuple):
                index_id, item = interpret_item_as_index_id_and_item(item)
            else:
                index_id = self.table.primary_key
        return index_id, item

    def _get_row_idxs_as_list(
        self,
        index_id: tuple,
        item,
        item_is_sequence: bool,
    ) -> list[int]:
        """
        Retrieve Table row indices for ``item`` as a list of integers.

        Parameters
        ----------
        index_id : tuple of str
            Identifier of the index to use.
        item : column element, list, ndarray, or slice
            Can be a value in the table index, a list/ndarray of such values, or a value
            slice (both endpoints are included).
        item_is_sequence : bool
            Whether ``item`` is a sequence (list or ndarray with ndim > 0) of values.

        Returns
        -------
        list of int
            List of row indices corresponding to the input item.

        Raises
        ------
        ValueError
            If the table has no indices.
        KeyError
            If no matches are found for a given key.
        """
        if len(self.indices) == 0:
            raise ValueError("Can only use TableLoc for a table with indices")

        index = self.indices[index_id]

        if isinstance(item, slice):
            start = None if item.start is None else (item.start,)
            stop = None if item.stop is None else (item.stop,)
            rows = index.range(start, stop)
        else:
            if not item_is_sequence:  # single element
                item = [item]
            # item should be a list or ndarray of values
            rows = []
            for value in item:
                ii = index.find(value if isinstance(value, tuple) else (value,))
                if len(ii) == 0:
                    raise KeyError(f"No matches found for key {value}")
                else:
                    rows.extend(ii)
        return rows

    def _get_row_idxs_as_list_or_int(self, item) -> list[int] | int:
        """Internal function to retrieve row indices for ``item`` as a list or int.

        See ``__getitem__`` for details on the input item.
        """
        # This handles ``tbl.loc[<item>]`` and ``tbl.loc[<key>, <item>]`` (and the same
        # for ``tbl.loc_indices``).
        index_id, item = self._get_index_id_and_item(item)

        item_is_sequence = (
            isinstance(item, np.ndarray) and item.ndim > 0
        ) or isinstance(item, list)

        # Short-circuit for case like tbl.loc[[]], returns tbl[[]]
        if item_is_sequence and len(item) == 0:
            return []

        row_idxs = self._get_row_idxs_as_list(index_id, item, item_is_sequence)
        # If ``item`` is a sequence of keys or a slice then always returns a list of
        # rows, where zero rows is OK. Otherwise check output and possibly return a
        # scalar.
        if not (item_is_sequence or isinstance(item, slice)) and len(row_idxs) == 1:
            row_idxs = row_idxs[0]

        return row_idxs

    def __getitem__(self, item) -> Table | Table.Row:
        """
        Retrieve Table rows by value slice.

        Parameters
        ----------
        item : column element, list, ndarray, slice or tuple
            Can be a value of the table primary index, a list/ndarray
            of such values, or a value slice (both endpoints are included).
            If a tuple is provided, the first element must be
            an index to use instead of the primary key, and the
            second element must be as above.

        Returns
        -------
        Table | Row
            A table slice or Row corresponding to the input item.
        """
        rows = self._get_row_idxs_as_list_or_int(item)
        return self.table[rows]

    def __setitem__(self, key, value):
        """
        Assign Table row's by value slice.

        Parameters
        ----------
        key : column element, list, ndarray, slice or tuple
              Can be a value of the table primary index, a list/ndarray
              of such values, or a value slice (both endpoints are included).
              If a tuple is provided, the first element must be
              an index to use instead of the primary key, and the
              second element must be as above.

        value : New values of the row elements.
                Can be a list of tuples/lists to update the row.
        """
        rows = self._get_row_idxs_as_list_or_int(key)
        if hasattr(rows, "__len__"):
            if len(rows) == len(value):
                for row, val in zip(rows, value):
                    self.table[row] = val
            else:
                raise ValueError(f"Right side should contain {len(rows)} values")
        else:
            self.table[rows] = value


class TableLocIndices(TableLoc):
    def __getitem__(self, item):
        """
        Retrieve Table row indices by value slice.

        Parameters
        ----------
        item : column element, list, ndarray, slice or tuple
               Can be a value of the table primary index, a list/ndarray
               of such values, or a value slice (both endpoints are included).
               If a tuple is provided, the first element must be
               an index to use instead of the primary key, and the
               second element must be as above.
        """
        rows = self._get_row_idxs_as_list_or_int(item)
        return rows


class TableILoc(TableLoc):
    """
    A variant of TableLoc allowing for row retrieval by
    indexed order rather than data values.

    Parameters
    ----------
    table : Table
        Indexed table to use
    """

    def __getitem__(self, item):
        if len(self.indices) == 0:
            raise ValueError("Can only use TableILoc for a table with indices")

        index_id, item = self._get_index_id_and_item(item)
        index = self.indices[index_id]
        rows = index.sorted_data()[item]
        table_slice = self.table[rows]

        if len(table_slice) == 0:  # no matches found
            raise IndexError(f"Invalid index for iloc: {item}")

        return table_slice



"""
This file tests the behavior of subclasses of Representation and Frames
"""

from copy import deepcopy

import astropy.coordinates
import astropy.units as u
from astropy.coordinates import ICRS, Latitude, Longitude
from astropy.coordinates.baseframe import RepresentationMapping, frame_transform_graph
from astropy.coordinates.representation import (
    REPRESENTATION_CLASSES,
    SphericalRepresentation,
    UnitSphericalRepresentation,
    get_reprdiff_cls_hash,
)
from astropy.coordinates.transformations import FunctionTransform

# Classes setup, borrowed from SunPy.

# Here we define the classes *inside* the tests to make sure that we can wipe
# the slate clean when the tests have finished running.


def setup_function(func):
    func.REPRESENTATION_CLASSES_ORIG = deepcopy(REPRESENTATION_CLASSES)


def teardown_function(func):
    REPRESENTATION_CLASSES.clear()
    REPRESENTATION_CLASSES.update(func.REPRESENTATION_CLASSES_ORIG)
    get_reprdiff_cls_hash.cache_clear()


def test_unit_representation_subclass():
    class Longitude180(Longitude):
        def __new__(cls, angle, unit=None, wrap_angle=180 * u.deg, **kwargs):
            return super().__new__(
                cls, angle, unit=unit, wrap_angle=wrap_angle, **kwargs
            )

    class UnitSphericalWrap180Representation(UnitSphericalRepresentation):
        attr_classes = {"lon": Longitude180, "lat": Latitude}

    class SphericalWrap180Representation(SphericalRepresentation):
        attr_classes = {"lon": Longitude180, "lat": Latitude, "distance": u.Quantity}

        _unit_representation = UnitSphericalWrap180Representation

    class MyFrame(ICRS):
        default_representation = SphericalWrap180Representation
        frame_specific_representation_info = {
            "spherical": [
                RepresentationMapping("lon", "ra"),
                RepresentationMapping("lat", "dec"),
            ]
        }
        frame_specific_representation_info[
            "unitsphericalwrap180"
        ] = frame_specific_representation_info["sphericalwrap180"] = (
            frame_specific_representation_info["spherical"]
        )

    @frame_transform_graph.transform(
        FunctionTransform, MyFrame, astropy.coordinates.ICRS
    )
    def myframe_to_icrs(myframe_coo, icrs):
        return icrs.realize_frame(myframe_coo._data)

    f = MyFrame(10 * u.deg, 10 * u.deg)
    assert isinstance(f._data, UnitSphericalWrap180Representation)
    assert isinstance(f.ra, Longitude180)

    g = f.transform_to(astropy.coordinates.ICRS())
    assert isinstance(g, astropy.coordinates.ICRS)
    assert isinstance(g._data, UnitSphericalWrap180Representation)

    frame_transform_graph.remove_transform(MyFrame, astropy.coordinates.ICRS, None)
